from importlib import metadata

import openapi_spec_validator as osv
import pytest
from flask.views import MethodView
from packaging.version import Version

from apiflask import APIBlueprint
from apiflask.security import APIKeyCookieAuth
from apiflask.security import APIKeyHeaderAuth
from apiflask.security import APIKeyQueryAuth
from apiflask.security import HTTPBasicAuth
from apiflask.security import HTTPTokenAuth
from apiflask.security import MultiAuth


werkzeug_version = Version(metadata.version('werkzeug'))


def test_auth_required(app, client):
    auth = HTTPBasicAuth()

    @auth.verify_password
    def verify_password(username, password):
        if username == 'foo' and password == 'bar':
            return {'user': 'foo'}
        elif username == 'bar' and password == 'foo':
            return {'user': 'bar'}
        elif username == 'baz' and password == 'baz':
            return {'user': 'baz'}

    @auth.get_user_roles
    def get_roles(user):
        if user['user'] == 'bar':
            return 'admin'
        elif user['user'] == 'baz':
            return 'moderator'
        return 'normal'

    @app.route('/foo')
    @app.auth_required(auth)
    def foo():
        return auth.current_user

    @app.route('/bar')
    @app.auth_required(auth, roles=['admin'])
    def bar():
        return auth.current_user

    @app.route('/baz')
    @app.auth_required(auth, roles=['admin', 'moderator'])
    def baz():
        return auth.current_user

    rv = client.get('/foo')
    assert rv.status_code == 401

    rv = client.get('/foo', headers={'Authorization': 'Basic Zm9vOmJhcg=='})
    assert rv.status_code == 200
    assert rv.json == {'user': 'foo'}

    rv = client.get('/bar', headers={'Authorization': 'Basic Zm9vOmJhcg=='})
    assert rv.status_code == 403

    rv = client.get('/foo', headers={'Authorization': 'Basic YmFyOmZvbw=='})
    assert rv.status_code == 200
    assert rv.json == {'user': 'bar'}

    rv = client.get('/bar', headers={'Authorization': 'Basic YmFyOmZvbw=='})
    assert rv.status_code == 200
    assert rv.json == {'user': 'bar'}

    rv = client.get('/baz', headers={'Authorization': 'Basic Zm9vOmJhcg=='})
    assert rv.status_code == 403

    rv = client.get('/baz', headers={'Authorization': 'Basic YmFyOmZvbw=='})
    assert rv.status_code == 200
    assert rv.json == {'user': 'bar'}

    rv = client.get('/baz', headers={'Authorization': 'Basic YmF6OmJheg=='})
    assert rv.status_code == 200
    assert rv.json == {'user': 'baz'}

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert 'BasicAuth' in rv.json['components']['securitySchemes']
    assert rv.json['components']['securitySchemes']['BasicAuth'] == {
        'scheme': 'basic',
        'type': 'http',
    }

    assert 'BasicAuth' in rv.json['paths']['/foo']['get']['security'][0]
    assert 'BasicAuth' in rv.json['paths']['/bar']['get']['security'][0]
    assert 'BasicAuth' in rv.json['paths']['/baz']['get']['security'][0]


def test_auth_required_with_methodview(app, client):
    auth = HTTPBasicAuth()

    @auth.verify_password
    def verify_password(username, password):
        if username == 'foo' and password == 'bar':
            return {'user': 'foo'}
        elif username == 'bar' and password == 'foo':
            return {'user': 'bar'}
        elif username == 'baz' and password == 'baz':
            return {'user': 'baz'}

    @auth.get_user_roles
    def get_roles(user):
        if user['user'] == 'bar':
            return 'admin'
        elif user['user'] == 'baz':
            return 'moderator'
        return 'normal'

    @app.route('/')
    class Foo(MethodView):
        @app.auth_required(auth)
        def get(self):
            return auth.current_user

        @app.auth_required(auth, roles=['admin'])
        def post(self):
            return auth.current_user

        @app.auth_required(auth, roles=['admin', 'moderator'])
        def delete(self):
            return auth.current_user

    rv = client.get('/')
    assert rv.status_code == 401

    rv = client.get('/', headers={'Authorization': 'Basic Zm9vOmJhcg=='})
    assert rv.status_code == 200
    assert rv.json == {'user': 'foo'}

    rv = client.post('/', headers={'Authorization': 'Basic Zm9vOmJhcg=='})
    assert rv.status_code == 403

    rv = client.get('/', headers={'Authorization': 'Basic YmFyOmZvbw=='})
    assert rv.status_code == 200
    assert rv.json == {'user': 'bar'}

    rv = client.post('/', headers={'Authorization': 'Basic YmFyOmZvbw=='})
    assert rv.status_code == 200
    assert rv.json == {'user': 'bar'}

    rv = client.delete('/', headers={'Authorization': 'Basic Zm9vOmJhcg=='})
    assert rv.status_code == 403

    rv = client.delete('/', headers={'Authorization': 'Basic YmFyOmZvbw=='})
    assert rv.status_code == 200
    assert rv.json == {'user': 'bar'}

    rv = client.delete('/', headers={'Authorization': 'Basic YmF6OmJheg=='})
    assert rv.status_code == 200
    assert rv.json == {'user': 'baz'}

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert 'BasicAuth' in rv.json['components']['securitySchemes']
    assert rv.json['components']['securitySchemes']['BasicAuth'] == {
        'scheme': 'basic',
        'type': 'http',
    }

    assert 'BasicAuth' in rv.json['paths']['/']['get']['security'][0]
    assert 'BasicAuth' in rv.json['paths']['/']['post']['security'][0]
    assert 'BasicAuth' in rv.json['paths']['/']['delete']['security'][0]


def test_auth_required_at_blueprint_before_request(app, client):
    bp = APIBlueprint('auth', __name__)
    no_auth_bp = APIBlueprint('no-auth', __name__)

    auth = HTTPTokenAuth()

    @bp.before_request
    @bp.auth_required(auth)
    def before():
        pass

    @bp.get('/foo')
    def foo():
        pass

    @bp.get('/bar')
    def bar():
        pass

    @bp.route('/baz')
    class Baz(MethodView):
        def get(self):
            pass

        def post(self):
            pass

    @no_auth_bp.get('/eggs')
    def eggs():
        return 'no auth'

    app.register_blueprint(bp)
    app.register_blueprint(no_auth_bp)

    rv = client.get('/foo')
    assert rv.status_code == 401
    rv = client.get('/bar')
    assert rv.status_code == 401
    rv = client.get('/baz')
    assert rv.status_code == 401
    rv = client.get('/eggs')
    assert rv.status_code == 200

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)

    assert 'auth' in app._auth_blueprints
    assert 'no-auth' not in app._auth_blueprints

    assert 'BearerAuth' in rv.json['components']['securitySchemes']
    assert rv.json['components']['securitySchemes']['BearerAuth'] == {
        'scheme': 'bearer',
        'type': 'http',
    }

    assert 'BearerAuth' in rv.json['paths']['/foo']['get']['security'][0]
    assert 'BearerAuth' in rv.json['paths']['/bar']['get']['security'][0]
    assert 'BearerAuth' in rv.json['paths']['/baz']['get']['security'][0]
    assert 'BearerAuth' in rv.json['paths']['/baz']['post']['security'][0]
    assert 'security' not in rv.json['paths']['/eggs']['get']


def test_lowercase_token_scheme_value(app, client):
    auth = HTTPTokenAuth(scheme='bearer')

    @app.route('/')
    @app.auth_required(auth)
    def index():
        pass

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)

    assert 'BearerAuth' in rv.json['components']['securitySchemes']
    assert 'BearerAuth' in rv.json['paths']['/']['get']['security'][0]


def test_auth_required_with_multiauth(app, client):
    basic_auth = HTTPBasicAuth()
    token_auth = HTTPTokenAuth()
    multi_auth = MultiAuth(basic_auth, token_auth)

    @basic_auth.verify_password
    def verify_password(username, password):
        if username == 'foo' and password == 'bar':
            return {'user': 'foo'}
        elif username == 'bar' and password == 'foo':
            return {'user': 'bar'}
        elif username == 'baz' and password == 'baz':
            return {'user': 'baz'}

    @basic_auth.get_user_roles
    def get_basic_roles(user):
        if user['user'] == 'bar':
            return 'admin'
        elif user['user'] == 'baz':
            return 'moderator'
        return 'normal'

    @token_auth.verify_token
    def verify_token(token):
        if token == 'foo_token':
            return {'user': 'foo'}
        elif token == 'bar_token':
            return {'user': 'bar'}
        elif token == 'baz_token':
            return {'user': 'baz'}

    @token_auth.get_user_roles
    def get_bearer_roles(user):
        if user['user'] == 'bar':
            return 'admin'
        elif user['user'] == 'baz':
            return 'moderator'
        return 'normal'

    @app.route('/foo')
    @app.auth_required(multi_auth)
    def foo():
        return multi_auth.current_user

    @app.route('/bar')
    @app.auth_required(multi_auth, roles=['admin'])
    def bar():
        return multi_auth.current_user

    @app.route('/baz')
    @app.auth_required(multi_auth, roles=['admin', 'moderator'])
    def baz():
        return multi_auth.current_user

    rv = client.get('/foo')
    assert rv.status_code == 401

    rv = client.get('/foo', headers={'Authorization': 'Basic Zm9vOmJhcg=='})
    assert rv.status_code == 200
    assert rv.json == {'user': 'foo'}

    rv = client.get('/bar', headers={'Authorization': 'Basic Zm9vOmJhcg=='})
    assert rv.status_code == 403

    rv = client.get('/foo', headers={'Authorization': 'Basic YmFyOmZvbw=='})
    assert rv.status_code == 200
    assert rv.json == {'user': 'bar'}

    rv = client.get('/bar', headers={'Authorization': 'Basic YmFyOmZvbw=='})
    assert rv.status_code == 200
    assert rv.json == {'user': 'bar'}

    rv = client.get('/baz', headers={'Authorization': 'Basic Zm9vOmJhcg=='})
    assert rv.status_code == 403

    rv = client.get('/baz', headers={'Authorization': 'Basic YmFyOmZvbw=='})
    assert rv.status_code == 200
    assert rv.json == {'user': 'bar'}

    rv = client.get('/baz', headers={'Authorization': 'Basic YmF6OmJheg=='})
    assert rv.status_code == 200
    assert rv.json == {'user': 'baz'}

    rv = client.get('/foo', headers={'Authorization': 'Bearer foo_token'})
    assert rv.status_code == 200
    assert rv.json == {'user': 'foo'}

    rv = client.get('/bar', headers={'Authorization': 'Bearer foo_token'})
    assert rv.status_code == 403

    rv = client.get('/foo', headers={'Authorization': 'Bearer bar_token'})
    assert rv.status_code == 200
    assert rv.json == {'user': 'bar'}

    rv = client.get('/bar', headers={'Authorization': 'Bearer bar_token'})
    assert rv.status_code == 200
    assert rv.json == {'user': 'bar'}

    rv = client.get('/baz', headers={'Authorization': 'Bearer foo_token'})
    assert rv.status_code == 403

    rv = client.get('/baz', headers={'Authorization': 'Bearer bar_token'})
    assert rv.status_code == 200
    assert rv.json == {'user': 'bar'}

    rv = client.get('/baz', headers={'Authorization': 'Bearer baz_token'})
    assert rv.status_code == 200
    assert rv.json == {'user': 'baz'}

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert 'BasicAuth' in rv.json['components']['securitySchemes']
    assert 'BearerAuth' in rv.json['components']['securitySchemes']

    assert rv.json['components']['securitySchemes']['BasicAuth'] == {
        'scheme': 'basic',
        'type': 'http',
    }
    assert rv.json['components']['securitySchemes']['BearerAuth'] == {
        'scheme': 'bearer',
        'type': 'http',
    }

    assert 'BasicAuth' in rv.json['paths']['/foo']['get']['security'][0]
    assert 'BasicAuth' in rv.json['paths']['/bar']['get']['security'][0]
    assert 'BasicAuth' in rv.json['paths']['/baz']['get']['security'][0]

    assert 'BearerAuth' in rv.json['paths']['/foo']['get']['security'][1]
    assert 'BearerAuth' in rv.json['paths']['/bar']['get']['security'][1]
    assert 'BearerAuth' in rv.json['paths']['/baz']['get']['security'][1]


@pytest.mark.skipif(
    werkzeug_version < Version('2.3.0'),
    reason='The first positional server_name parameter to set_cookie and delete_cookie is deprecated. Use the domain parameter instead.',  # noqa: E501
)
def test_apikey_auth_required_with_multiauth(app, client):
    header_auth = APIKeyHeaderAuth(name='APIKeyHeaderAuth')
    cookie_auth = APIKeyCookieAuth(name='APIKeyCookieAuth')
    query_auth = APIKeyQueryAuth(name='APIKeyQueryAuth')
    multi_auth = MultiAuth(header_auth, cookie_auth, query_auth)

    @header_auth.verify_token
    @cookie_auth.verify_token
    @query_auth.verify_token
    def verify_token(token):
        if token == 'foo_token':
            return {'user': 'foo'}
        elif token == 'bar_token':
            return {'user': 'bar'}
        elif token == 'baz_token':
            return {'user': 'baz'}

    @header_auth.get_user_roles
    @cookie_auth.get_user_roles
    @query_auth.get_user_roles
    def get_bearer_roles(user):
        if user['user'] == 'bar':
            return 'admin'
        elif user['user'] == 'baz':
            return 'moderator'
        return 'normal'

    @app.route('/foo')
    @app.auth_required(multi_auth)
    def foo():
        return multi_auth.current_user

    @app.route('/bar')
    @app.auth_required(multi_auth, roles=['admin'])
    def bar():
        return multi_auth.current_user

    @app.route('/baz')
    @app.auth_required(multi_auth, roles=['admin', 'moderator'])
    def baz():
        return multi_auth.current_user

    rv = client.get('/foo')
    assert rv.status_code == 401

    rv = client.get('/bar', headers={'X-API-Key': 'foo_token'})
    assert rv.status_code == 403

    rv = client.get('/foo', headers={'X-API-Key': 'bar_token'})
    assert rv.status_code == 200
    assert rv.json == {'user': 'bar'}

    rv = client.get('/bar', headers={'X-API-Key': 'bar_token'})
    assert rv.status_code == 200
    assert rv.json == {'user': 'bar'}

    rv = client.get('/baz', headers={'X-API-Key': 'foo_token'})
    assert rv.status_code == 403

    rv = client.get('/baz', headers={'X-API-Key': 'bar_token'})
    assert rv.status_code == 200
    assert rv.json == {'user': 'bar'}

    rv = client.get('/baz', headers={'X-API-Key': 'baz_token'})
    assert rv.status_code == 200
    assert rv.json == {'user': 'baz'}

    client.set_cookie(key='X-API-Key', value='foo_token')
    rv = client.get('/bar')
    assert rv.status_code == 403
    client.delete_cookie('X-API-Key')

    client.set_cookie(key='X-API-Key', value='bar_token')
    rv = client.get('/foo')
    assert rv.status_code == 200
    assert rv.json == {'user': 'bar'}
    client.delete_cookie('X-API-Key')

    client.set_cookie(key='X-API-Key', value='bar_token')
    rv = client.get('/bar')
    assert rv.status_code == 200
    assert rv.json == {'user': 'bar'}
    client.delete_cookie('X-API-Key')

    client.set_cookie(key='X-API-Key', value='foo_token')
    rv = client.get('/baz')
    assert rv.status_code == 403
    client.delete_cookie('X-API-Key')

    client.set_cookie(key='X-API-Key', value='bar_token')
    rv = client.get('/baz')
    assert rv.status_code == 200
    assert rv.json == {'user': 'bar'}
    client.delete_cookie('X-API-Key')

    client.set_cookie(key='X-API-Key', value='baz_token')
    rv = client.get('/baz')
    assert rv.status_code == 200
    assert rv.json == {'user': 'baz'}
    client.delete_cookie('X-API-Key')

    rv = client.get('/bar', query_string={'X-API-Key': 'foo_token'})
    assert rv.status_code == 403

    rv = client.get('/foo', query_string={'X-API-Key': 'bar_token'})
    assert rv.status_code == 200
    assert rv.json == {'user': 'bar'}

    rv = client.get('/bar', query_string={'X-API-Key': 'bar_token'})
    assert rv.status_code == 200
    assert rv.json == {'user': 'bar'}

    rv = client.get('/baz', query_string={'X-API-Key': 'foo_token'})
    assert rv.status_code == 403

    rv = client.get('/baz', query_string={'X-API-Key': 'bar_token'})
    assert rv.status_code == 200
    assert rv.json == {'user': 'bar'}

    rv = client.get('/baz', query_string={'X-API-Key': 'baz_token'})
    assert rv.status_code == 200
    assert rv.json == {'user': 'baz'}

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert 'APIKeyHeaderAuth' in rv.json['components']['securitySchemes']
    assert 'APIKeyCookieAuth' in rv.json['components']['securitySchemes']
    assert 'APIKeyQueryAuth' in rv.json['components']['securitySchemes']

    assert rv.json['components']['securitySchemes']['APIKeyHeaderAuth'] == {
        'type': 'apiKey',
        'name': 'X-API-Key',
        'in': 'header',
    }
    assert rv.json['components']['securitySchemes']['APIKeyCookieAuth'] == {
        'type': 'apiKey',
        'name': 'X-API-Key',
        'in': 'cookie',
    }
    assert rv.json['components']['securitySchemes']['APIKeyQueryAuth'] == {
        'type': 'apiKey',
        'name': 'X-API-Key',
        'in': 'query',
    }

    assert 'APIKeyHeaderAuth' in rv.json['paths']['/foo']['get']['security'][0]
    assert 'APIKeyHeaderAuth' in rv.json['paths']['/bar']['get']['security'][0]
    assert 'APIKeyHeaderAuth' in rv.json['paths']['/baz']['get']['security'][0]

    assert 'APIKeyCookieAuth' in rv.json['paths']['/foo']['get']['security'][1]
    assert 'APIKeyCookieAuth' in rv.json['paths']['/bar']['get']['security'][1]
    assert 'APIKeyCookieAuth' in rv.json['paths']['/baz']['get']['security'][1]

    assert 'APIKeyQueryAuth' in rv.json['paths']['/foo']['get']['security'][2]
    assert 'APIKeyQueryAuth' in rv.json['paths']['/bar']['get']['security'][2]
    assert 'APIKeyQueryAuth' in rv.json['paths']['/baz']['get']['security'][2]


@pytest.mark.skipif(
    werkzeug_version >= Version('2.3.0'), reason='Old usage of set_cookie and delete_cookie.'
)
def test_apikey_auth_required_with_multiauth_with_old_version_werkzeug(app, client):
    header_auth = APIKeyHeaderAuth(name='APIKeyHeaderAuth')
    cookie_auth = APIKeyCookieAuth(name='APIKeyCookieAuth')
    query_auth = APIKeyQueryAuth(name='APIKeyQueryAuth')
    multi_auth = MultiAuth(header_auth, cookie_auth, query_auth)

    @header_auth.verify_token
    @cookie_auth.verify_token
    @query_auth.verify_token
    def verify_token(token):
        if token == 'foo_token':
            return {'user': 'foo'}
        elif token == 'bar_token':
            return {'user': 'bar'}
        elif token == 'baz_token':
            return {'user': 'baz'}

    @header_auth.get_user_roles
    @cookie_auth.get_user_roles
    @query_auth.get_user_roles
    def get_bearer_roles(user):
        if user['user'] == 'bar':
            return 'admin'
        elif user['user'] == 'baz':
            return 'moderator'
        return 'normal'

    @app.route('/foo')
    @app.auth_required(multi_auth)
    def foo():
        return multi_auth.current_user

    @app.route('/bar')
    @app.auth_required(multi_auth, roles=['admin'])
    def bar():
        return multi_auth.current_user

    @app.route('/baz')
    @app.auth_required(multi_auth, roles=['admin', 'moderator'])
    def baz():
        return multi_auth.current_user

    rv = client.get('/foo')
    assert rv.status_code == 401

    rv = client.get('/bar', headers={'X-API-Key': 'foo_token'})
    assert rv.status_code == 403

    rv = client.get('/foo', headers={'X-API-Key': 'bar_token'})
    assert rv.status_code == 200
    assert rv.json == {'user': 'bar'}

    rv = client.get('/bar', headers={'X-API-Key': 'bar_token'})
    assert rv.status_code == 200
    assert rv.json == {'user': 'bar'}

    rv = client.get('/baz', headers={'X-API-Key': 'foo_token'})
    assert rv.status_code == 403

    rv = client.get('/baz', headers={'X-API-Key': 'bar_token'})
    assert rv.status_code == 200
    assert rv.json == {'user': 'bar'}

    rv = client.get('/baz', headers={'X-API-Key': 'baz_token'})
    assert rv.status_code == 200
    assert rv.json == {'user': 'baz'}

    client.set_cookie(key='X-API-Key', value='foo_token', server_name='localhost')
    rv = client.get('/bar')
    assert rv.status_code == 403
    client.delete_cookie(key='X-API-Key', server_name='localhost')

    client.set_cookie(key='X-API-Key', value='bar_token', server_name='localhost')
    rv = client.get('/foo')
    assert rv.status_code == 200
    assert rv.json == {'user': 'bar'}
    client.delete_cookie(key='X-API-Key', server_name='localhost')

    client.set_cookie(key='X-API-Key', value='bar_token', server_name='localhost')
    rv = client.get('/bar')
    assert rv.status_code == 200
    assert rv.json == {'user': 'bar'}
    client.delete_cookie(key='X-API-Key', server_name='localhost')

    client.set_cookie(key='X-API-Key', value='foo_token', server_name='localhost')
    rv = client.get('/baz')
    assert rv.status_code == 403
    client.delete_cookie(key='X-API-Key', server_name='localhost')

    client.set_cookie(key='X-API-Key', value='bar_token', server_name='localhost')
    rv = client.get('/baz')
    assert rv.status_code == 200
    assert rv.json == {'user': 'bar'}
    client.delete_cookie(key='X-API-Key', server_name='localhost')

    client.set_cookie(key='X-API-Key', value='baz_token', server_name='localhost')
    rv = client.get('/baz')
    assert rv.status_code == 200
    assert rv.json == {'user': 'baz'}
    client.delete_cookie(key='X-API-Key', server_name='localhost')

    rv = client.get('/bar', query_string={'X-API-Key': 'foo_token'})
    assert rv.status_code == 403

    rv = client.get('/foo', query_string={'X-API-Key': 'bar_token'})
    assert rv.status_code == 200
    assert rv.json == {'user': 'bar'}

    rv = client.get('/bar', query_string={'X-API-Key': 'bar_token'})
    assert rv.status_code == 200
    assert rv.json == {'user': 'bar'}

    rv = client.get('/baz', query_string={'X-API-Key': 'foo_token'})
    assert rv.status_code == 403

    rv = client.get('/baz', query_string={'X-API-Key': 'bar_token'})
    assert rv.status_code == 200
    assert rv.json == {'user': 'bar'}

    rv = client.get('/baz', query_string={'X-API-Key': 'baz_token'})
    assert rv.status_code == 200
    assert rv.json == {'user': 'baz'}

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert 'APIKeyHeaderAuth' in rv.json['components']['securitySchemes']
    assert 'APIKeyCookieAuth' in rv.json['components']['securitySchemes']
    assert 'APIKeyQueryAuth' in rv.json['components']['securitySchemes']

    assert rv.json['components']['securitySchemes']['APIKeyHeaderAuth'] == {
        'type': 'apiKey',
        'name': 'X-API-Key',
        'in': 'header',
    }
    assert rv.json['components']['securitySchemes']['APIKeyCookieAuth'] == {
        'type': 'apiKey',
        'name': 'X-API-Key',
        'in': 'cookie',
    }
    assert rv.json['components']['securitySchemes']['APIKeyQueryAuth'] == {
        'type': 'apiKey',
        'name': 'X-API-Key',
        'in': 'query',
    }

    assert 'APIKeyHeaderAuth' in rv.json['paths']['/foo']['get']['security'][0]
    assert 'APIKeyHeaderAuth' in rv.json['paths']['/bar']['get']['security'][0]
    assert 'APIKeyHeaderAuth' in rv.json['paths']['/baz']['get']['security'][0]

    assert 'APIKeyCookieAuth' in rv.json['paths']['/foo']['get']['security'][1]
    assert 'APIKeyCookieAuth' in rv.json['paths']['/bar']['get']['security'][1]
    assert 'APIKeyCookieAuth' in rv.json['paths']['/baz']['get']['security'][1]

    assert 'APIKeyQueryAuth' in rv.json['paths']['/foo']['get']['security'][2]
    assert 'APIKeyQueryAuth' in rv.json['paths']['/bar']['get']['security'][2]
    assert 'APIKeyQueryAuth' in rv.json['paths']['/baz']['get']['security'][2]



import sys

import openapi_spec_validator as osv
import pytest

from apiflask import APIFlask

# Skip all tests in this module if pydantic is not available
pydantic = pytest.importorskip('pydantic')
BaseModel = pydantic.BaseModel


class UserModel(BaseModel):
    id: int
    name: str
    email: str


class UserCreateModel(BaseModel):
    name: str
    email: str


class TestPydanticIntegration:
    """Test Pydantic model integration with APIFlask."""

    def test_input_decorator_with_pydantic_model(self):
        """Test @app.input decorator with Pydantic model."""
        app = APIFlask(__name__)

        @app.post('/users')
        @app.input(UserCreateModel, location='json')
        def create_user(json_data):
            assert isinstance(json_data, UserCreateModel)
            assert json_data.name == 'John Doe'
            assert json_data.email == 'john@example.com'
            return {'message': 'User created'}

        with app.test_client() as client:
            response = client.post('/users', json={'name': 'John Doe', 'email': 'john@example.com'})
            assert response.status_code == 200

    def test_output_decorator_with_pydantic_model(self):
        """Test @app.output decorator with Pydantic model."""
        app = APIFlask(__name__)

        @app.get('/users/<int:user_id>')
        @app.output(UserModel)
        def get_user(user_id):
            return UserModel(id=user_id, name='John Doe', email='john@example.com')

        with app.test_client() as client:
            response = client.get('/users/1')
            assert response.status_code == 200
            data = response.get_json()
            assert data['id'] == 1
            assert data['name'] == 'John Doe'
            assert data['email'] == 'john@example.com'

    def test_input_validation_error(self):
        """Test validation error with Pydantic model."""
        app = APIFlask(__name__)

        @app.post('/users')
        @app.input(UserCreateModel, location='json')
        def create_user(json_data):
            return {'message': 'User created'}

        with app.test_client() as client:
            # Missing required field
            response = client.post('/users', json={'name': 'John Doe'})
            assert response.status_code == app.config['VALIDATION_ERROR_STATUS_CODE']
            data = response.get_json()
            assert 'message' in data
            assert 'detail' in data

    def test_pydantic_openapi_schema_generation(self):
        """Test OpenAPI schema generation for Pydantic models."""
        app = APIFlask(__name__)

        @app.post('/users')
        @app.input(UserCreateModel, location='json')
        @app.output(UserModel)
        def create_user(json_data):
            return UserModel(id=1, name=json_data.name, email=json_data.email)

        with app.test_client() as client:
            rv = client.get('/openapi.json')
            assert rv.status_code == 200

            # Validate OpenAPI spec
            osv.validate(rv.json)

            spec = rv.get_json()
            assert 'components' in spec or 'definitions' in spec
            # Check that the schema is generated
            assert 'paths' in spec
            assert '/users' in spec['paths']

    def test_pydantic_models_in_components_schemas(self):
        """Test that Pydantic models are correctly registered in components/schemas."""
        app = APIFlask(__name__)

        @app.post('/users')
        @app.input(UserCreateModel, location='json')
        @app.output(UserModel, status_code=201)
        def create_user(json_data):
            return UserModel(id=1, name=json_data.name, email=json_data.email)

        @app.get('/users/<int:user_id>')
        @app.output(UserModel)
        def get_user(user_id):
            return UserModel(id=user_id, name='John Doe', email='john@example.com')

        with app.test_client() as client:
            # Fetch the OpenAPI spec
            rv = client.get('/openapi.json')
            assert rv.status_code == 200

            # Validate OpenAPI spec
            osv.validate(rv.json)

            spec = rv.get_json()

            # Check that both Pydantic models are in components/schemas
            assert 'components' in spec
            assert 'schemas' in spec['components']
            schemas = spec['components']['schemas']

            # UserModel and UserCreateModel should be registered
            assert 'UserModel' in schemas
            assert 'UserCreateModel' in schemas

            # Check UserModel schema structure
            user_model_schema = schemas['UserModel']
            assert 'properties' in user_model_schema
            assert 'id' in user_model_schema['properties']
            assert 'name' in user_model_schema['properties']
            assert 'email' in user_model_schema['properties']
            assert user_model_schema['properties']['id']['type'] == 'integer'
            assert user_model_schema['properties']['name']['type'] == 'string'
            assert user_model_schema['properties']['email']['type'] == 'string'

            # Check UserCreateModel schema structure
            user_create_schema = schemas['UserCreateModel']
            assert 'properties' in user_create_schema
            assert 'name' in user_create_schema['properties']
            assert 'email' in user_create_schema['properties']
            assert 'id' not in user_create_schema['properties']  # Should not have id

    def test_pydantic_schema_references_in_paths(self):
        """Test that schema references in paths use correct names, not class representations."""
        app = APIFlask(__name__)

        @app.post('/users')
        @app.input(UserCreateModel, location='json')
        @app.output(UserModel, status_code=201)
        def create_user(json_data):
            return UserModel(id=1, name=json_data.name, email=json_data.email)

        @app.get('/users/<int:user_id>')
        @app.output(UserModel)
        def get_user(user_id):
            return UserModel(id=user_id, name='John Doe', email='john@example.com')

        with app.test_client() as client:
            rv = client.get('/openapi.json')
            assert rv.status_code == 200

            # Validate OpenAPI spec
            osv.validate(rv.json)

            spec = rv.get_json()

            # Check POST /users requestBody reference
            post_path = spec['paths']['/users']['post']
            assert 'requestBody' in post_path
            request_schema = post_path['requestBody']['content']['application/json']['schema']
            assert '$ref' in request_schema
            # Should be a proper reference, not a class representation
            assert request_schema['$ref'] == '#/components/schemas/UserCreateModel'
            assert '<class' not in request_schema['$ref']  # Should NOT contain <class ...>

            # Check POST /users response reference
            post_response = post_path['responses']['201']['content']['application/json']['schema']
            assert '$ref' in post_response
            assert post_response['$ref'] == '#/components/schemas/UserModel'
            assert '<class' not in post_response['$ref']

            # Check GET /users/{user_id} response reference
            get_path = spec['paths']['/users/{user_id}']['get']
            get_response = get_path['responses']['200']['content']['application/json']['schema']
            assert '$ref' in get_response
            assert get_response['$ref'] == '#/components/schemas/UserModel'
            assert '<class' not in get_response['$ref']

    def test_mixed_schemas_support(self):
        """Test that both marshmallow and Pydantic schemas can coexist."""
        from apiflask import Schema
        from apiflask.fields import String, Integer

        app = APIFlask(__name__)

        # Marshmallow schema
        class UserMarshmallowSchema(Schema):
            id = Integer()
            name = String()

        @app.get('/users/marshmallow/<int:user_id>')
        @app.output(UserMarshmallowSchema)
        def get_user_marshmallow(user_id):
            return {'id': user_id, 'name': 'John Doe'}

        @app.get('/users/pydantic/<int:user_id>')
        @app.output(UserModel)
        def get_user_pydantic(user_id):
            return UserModel(id=user_id, name='John Doe', email='john@example.com')

        with app.test_client() as client:
            # Validate OpenAPI spec with mixed schemas
            rv = client.get('/openapi.json')
            assert rv.status_code == 200
            osv.validate(rv.json)

            # Test marshmallow endpoint
            response = client.get('/users/marshmallow/1')
            assert response.status_code == 200

            # Test pydantic endpoint
            response = client.get('/users/pydantic/1')
            assert response.status_code == 200

    def test_query_parameters_with_pydantic(self):
        """Test query parameter validation with Pydantic."""

        class QueryModel(BaseModel):
            page: int = 1
            per_page: int = 10
            search: str = ''

        app = APIFlask(__name__)

        @app.get('/users')
        @app.input(QueryModel, location='query')
        def list_users(query_data):
            assert isinstance(query_data, QueryModel)
            return {
                'page': query_data.page,
                'per_page': query_data.per_page,
                'search': query_data.search,
            }

        with app.test_client() as client:
            # Validate OpenAPI spec
            rv = client.get('/openapi.json')
            assert rv.status_code == 200
            osv.validate(rv.json)

            response = client.get('/users?page=2&per_page=20&search=john')
            assert response.status_code == 200
            data = response.get_json()
            assert data['page'] == 2
            assert data['per_page'] == 20
            assert data['search'] == 'john'

    def test_headers_with_pydantic(self):
        """Test headers validation with Pydantic."""

        class HeaderModel(BaseModel):
            x_token: str = 'default'
            x_version: str = '1.0'

        app = APIFlask(__name__)

        @app.get('/protected')
        @app.input(HeaderModel, location='headers')
        def protected_route(headers_data):
            assert isinstance(headers_data, HeaderModel)
            return {
                'token': headers_data.x_token,
                'version': headers_data.x_version,
            }

        with app.test_client() as client:
            response = client.get(
                '/protected', headers={'X-Token': 'secret123', 'X-Version': '2.0'}
            )
            assert response.status_code == 200
            data = response.get_json()
            assert data['token'] == 'secret123'
            assert data['version'] == '2.0'

            rv = client.get('/openapi.json')
            assert rv.status_code == 200
            osv.validate(rv.json)

    def test_cookies_with_pydantic(self):
        """Test cookies validation with Pydantic."""

        class CookieModel(BaseModel):
            session_id: str = 'default'
            theme: str = 'light'

        app = APIFlask(__name__)

        @app.get('/dashboard')
        @app.input(CookieModel, location='cookies')
        def dashboard(cookies_data):
            assert isinstance(cookies_data, CookieModel)
            return {
                'session_id': cookies_data.session_id,
                'theme': cookies_data.theme,
            }

        with app.test_client() as client:
            # Flask 2.0-2.2 requires server_name as positional arg
            # Flask 2.3+ uses domain parameter instead
            import inspect

            sig = inspect.signature(client.set_cookie)
            if 'server_name' in sig.parameters and sig.parameters['server_name'].kind in (
                inspect.Parameter.POSITIONAL_ONLY,
                inspect.Parameter.POSITIONAL_OR_KEYWORD,
            ):
                # Old API (Flask 2.0-2.2)
                client.set_cookie('localhost', 'session_id', 'abc123')
                client.set_cookie('localhost', 'theme', 'dark')
            else:
                # New API (Flask 2.3+)
                client.set_cookie(key='session_id', value='abc123', domain='localhost')
                client.set_cookie(key='theme', value='dark', domain='localhost')
            response = client.get('/dashboard')
            assert response.status_code == 200
            data = response.get_json()
            assert data['session_id'] == 'abc123'
            assert data['theme'] == 'dark'

            rv = client.get('/openapi.json')
            assert rv.status_code == 200
            osv.validate(rv.json)

    def test_path_parameters_with_pydantic(self):
        """Test path parameters validation with Pydantic."""

        class PathModel(BaseModel):
            user_id: int
            action: str

        app = APIFlask(__name__)

        @app.get('/users/<int:user_id>/<action>')
        @app.input(PathModel, location='path')
        def user_action(user_id, action, path_data):
            assert isinstance(path_data, PathModel)
            return {
                'user_id': path_data.user_id,
                'action': path_data.action,
            }

        with app.test_client() as client:
            # Validate OpenAPI spec
            rv = client.get('/openapi.json')
            assert rv.status_code == 200
            osv.validate(rv.json)

            response = client.get('/users/123/delete')
            assert response.status_code == 200
            data = response.get_json()
            assert data['user_id'] == 123
            assert data['action'] == 'delete'

    def test_view_args_location_alias(self):
        """Test that 'view_args' location works as alias for 'path'."""

        class PathModel(BaseModel):
            item_id: int

        app = APIFlask(__name__)

        @app.get('/items/<int:item_id>')
        @app.input(PathModel, location='view_args')
        def get_item(item_id, view_args_data):
            assert isinstance(view_args_data, PathModel)
            return {'item_id': view_args_data.item_id}

        with app.test_client() as client:
            response = client.get('/items/456')
            assert response.status_code == 200
            data = response.get_json()
            assert data['item_id'] == 456

            rv = client.get('/openapi.json')
            assert rv.status_code == 200
            osv.validate(rv.json)

    def test_list_type_with_typing_List(self):
        """Test list type support with typing.List[Model]."""
        from typing import List

        app = APIFlask(__name__)

        @app.get('/users')
        @app.output(List[UserModel])
        def get_users():
            return [
                UserModel(id=1, name='Alice', email='alice@example.com'),
                UserModel(id=2, name='Bob', email='bob@example.com'),
            ]

        with app.test_client() as client:
            # Test response
            response = client.get('/users')
            assert response.status_code == 200
            data = response.get_json()
            assert isinstance(data, list)
            assert len(data) == 2
            assert data[0]['id'] == 1
            assert data[0]['name'] == 'Alice'
            assert data[1]['id'] == 2
            assert data[1]['name'] == 'Bob'

            # Test OpenAPI spec
            rv = client.get('/openapi.json')
            assert rv.status_code == 200
            osv.validate(rv.json)

            spec = rv.get_json()
            response_schema = spec['paths']['/users']['get']['responses']['200']['content'][
                'application/json'
            ]['schema']
            # Should be array type with items referencing UserModel
            assert response_schema['type'] == 'array'
            assert 'items' in response_schema
            assert response_schema['items']['$ref'] == '#/components/schemas/UserModel'

    @pytest.mark.skipif(sys.version_info < (3, 9), reason='Native list syntax requires Python 3.9+')
    def test_list_type_with_native_list(self):
        """Test list type support with native list[Model] (Python 3.9+)."""
        app = APIFlask(__name__)

        @app.get('/users')
        @app.output(list[UserModel])
        def get_users():
            return [
                {'id': 1, 'name': 'Alice', 'email': 'alice@example.com'},
                {'id': 2, 'name': 'Bob', 'email': 'bob@example.com'},
            ]

        with app.test_client() as client:
            # Test response
            response = client.get('/users')
            assert response.status_code == 200
            data = response.get_json()
            assert isinstance(data, list)
            assert len(data) == 2
            assert data[0]['id'] == 1
            assert data[1]['id'] == 2

            # Test OpenAPI spec
            rv = client.get('/openapi.json')
            assert rv.status_code == 200
            osv.validate(rv.json)

            spec = rv.get_json()
            response_schema = spec['paths']['/users']['get']['responses']['200']['content'][
                'application/json'
            ]['schema']
            # Should be array type
            assert response_schema['type'] == 'array'
            assert 'items' in response_schema
            assert response_schema['items']['$ref'] == '#/components/schemas/UserModel'

    @pytest.mark.skipif(sys.version_info < (3, 9), reason='Native list syntax requires Python 3.9+')
    def test_list_type_with_dict_response(self):
        """Test list type with dict responses instead of model instances."""
        app = APIFlask(__name__)

        @app.get('/users')
        @app.output(list[UserModel])
        def get_users():
            # Return list of dicts - should be validated and serialized
            return [
                {'id': 1, 'name': 'Alice', 'email': 'alice@example.com'},
                {'id': 2, 'name': 'Bob', 'email': 'bob@example.com'},
            ]

        with app.test_client() as client:
            response = client.get('/users')
            assert response.status_code == 200
            data = response.get_json()
            assert isinstance(data, list)
            assert len(data) == 2
            assert all('id' in item and 'name' in item and 'email' in item for item in data)

    def test_mixed_list_and_single_endpoints(self):
        """Test that list and single endpoints work correctly together."""
        from typing import List

        app = APIFlask(__name__)

        @app.get('/users')
        @app.output(List[UserModel])
        def list_users():
            return [
                UserModel(id=1, name='Alice', email='alice@example.com'),
                UserModel(id=2, name='Bob', email='bob@example.com'),
            ]

        @app.get('/users/<int:user_id>')
        @app.output(UserModel)
        def get_user(user_id):
            return UserModel(id=user_id, name='Alice', email='alice@example.com')

        @app.post('/users')
        @app.input(UserCreateModel, location='json')
        @app.output(UserModel, status_code=201)
        def create_user(json_data):
            return UserModel(id=1, name=json_data.name, email=json_data.email)

        with app.test_client() as client:
            # Test list endpoint
            response = client.get('/users')
            assert response.status_code == 200
            data = response.get_json()
            assert isinstance(data, list)
            assert len(data) == 2

            # Test single endpoint
            response = client.get('/users/1')
            assert response.status_code == 200
            data = response.get_json()
            assert isinstance(data, dict)
            assert data['id'] == 1

            # Test OpenAPI spec
            rv = client.get('/openapi.json')
            assert rv.status_code == 200
            osv.validate(rv.json)

            spec = rv.get_json()

            # List endpoint should have array schema
            list_schema = spec['paths']['/users']['get']['responses']['200']['content'][
                'application/json'
            ]['schema']
            assert list_schema['type'] == 'array'
            assert list_schema['items']['$ref'] == '#/components/schemas/UserModel'

            # Single endpoint should have object schema
            single_schema = spec['paths']['/users/{user_id}']['get']['responses']['200']['content'][
                'application/json'
            ]['schema']
            assert '$ref' in single_schema
            assert single_schema['$ref'] == '#/components/schemas/UserModel'
            assert 'type' not in single_schema  # Should not be wrapped in array



from __future__ import annotations

import inspect
import re
import typing as t
import warnings
from functools import wraps

from apispec import APISpec
from apispec import BasePlugin
from apispec.ext.marshmallow import MarshmallowPlugin
from flask import Blueprint
from flask import Flask
from flask import has_request_context
from flask import json
from flask import jsonify
from flask import render_template_string
from flask import request
from flask import url_for
from flask.config import ConfigAttribute
from flask.wrappers import Response

with warnings.catch_warnings():
    warnings.simplefilter('ignore')
    from flask_marshmallow import fields

    try:
        from flask_marshmallow import sqla
    except ImportError:
        sqla = None  # type: ignore

from werkzeug.exceptions import HTTPException as WerkzeugHTTPException

from .exceptions import HTTPError
from .exceptions import _bad_schema_message
from .helpers import get_reason_phrase
from .route import route_patch
from .schemas import Schema
from .schemas import FileSchema
from .schemas import EmptySchema
from .schema_adapters import registry
from .security import MultiAuth
from .security import _AuthBase
from .types import SecurityScheme
from .types import ResponseReturnValueType, ResponsesType
from .types import ViewFuncType
from .types import ErrorCallbackType
from .types import SpecCallbackType
from .types import SchemaType
from .types import HTTPAuthType
from .types import TagsType
from .types import OpenAPISchemaType
from .openapi import default_bypassed_endpoints
from .openapi import default_response
from .openapi import get_tag
from .openapi import get_operation_tags
from .openapi import get_path_summary
from .openapi import get_argument
from .openapi import get_security_and_security_schemes
from .openapi_adapters import get_unique_schema_name
from .openapi_adapters import openapi_helper
from .openapi_adapters import extract_pydantic_defs
from .ui_templates import ui_templates
from .ui_templates import swagger_ui_oauth2_redirect_template
from .scaffold import APIScaffold


@route_patch
class APIFlask(APIScaffold, Flask):
    """The `Flask` object with some web API support.

    Examples:

    ```python
    from apiflask import APIFlask

    app = APIFlask(__name__)
    ```

    Attributes:
        openapi_version: The version of OpenAPI Specification (openapi.openapi).
            This attribute can also be configured from the config with the
            `OPENAPI_VERSION` configuration key. Defaults to `'3.0.3'`.
        servers: The servers information of the API (openapi.servers), accepts
            multiple server dicts. Example value:

            ```python
            app.servers = [
                {
                    'name': 'Production Server',
                    'url': 'http://api.example.com'
                }
            ]
            ```

            This attribute can also be configured from the config with the
            `SERVERS` configuration key. Defaults to `None`.
        tags: The list of tags of the OpenAPI spec documentation (openapi.tags),
            accepts a list of dicts. You can also pass a simple list contains the
            tag name:

            ```python
            app.tags = ['foo', 'bar', 'baz']
            ```

            A standard OpenAPI tags list will look like this:

            ```python
            app.tags = [
                {'name': 'foo', 'description': 'The description of foo'},
                {'name': 'bar', 'description': 'The description of bar'},
                {'name': 'baz', 'description': 'The description of baz'}
            ]
            ```

            If not set, the blueprint names will be used as tags.

            This attribute can also be configured from the config with the
            `TAGS` configuration key. Defaults to `None`.
        external_docs: The external documentation information of the API
            (openapi.externalDocs). Example:

            ```python
            app.external_docs = {
                'description': 'Find more info here',
                'url': 'http://docs.example.com'
            }
            ```

            This attribute can also be configured from the config with the
            `EXTERNAL_DOCS` configuration key. Defaults to `None`.
        info: The info object (openapi.info), it accepts a dict contains following info fields:
            `description`, `termsOfService`, `contact`, `license`. You can use separate
            configuration variables to overwrite this dict. Example:

            ```python
            app.info = {
                'description': '...',
                'termsOfService': 'http://example.com',
                'contact': {
                    'name': 'API Support',
                    'url': 'http://www.example.com/support',
                    'email': 'support@example.com'
                },
                'license': {
                    'name': 'Apache 2.0',
                    'url': 'http://www.apache.org/licenses/LICENSE-2.0.html'
                }
            }
            ```

        description: The description of the API (openapi.info.description).
            This attribute can also be configured from the config with the
            `DESCRIPTION` configuration key. Defaults to `None`.
        contact: The contact information of the API (openapi.info.contact). Example:

            ```python
            app.contact = {
                'name': 'API Support',
                'url': 'http://www.example.com/support',
                'email': 'support@example.com'
            }
            ```

            This attribute can also be configured from the config with the
            `CONTACT` configuration key. Defaults to `None`.
        license: The license of the API (openapi.info.license). Example:

            ```python
            app.license = {
                'name': 'Apache 2.0',
                'url': 'http://www.apache.org/licenses/LICENSE-2.0.html'
            }
            ```

            This attribute can also be configured from the config with the
            `LICENSE` configuration key. Defaults to `None`.
        terms_of_service: The terms of service URL of the API
            (openapi.info.termsOfService). Example:

            ```python
            app.terms_of_service = 'http://example.com/terms/'
            ```

            This attribute can also be configured from the config with the
            `TERMS_OF_SERVICE` configuration key. Defaults to `None`.
        security_schemes: The security schemes of the API
            (openapi.components.securitySchemes). Example:

            ```python
            app.security_schemes = [
                'ApiKeyAuth': {
                    'type': 'apiKey',
                    'in': 'header',
                    'name': 'X-API-Key'
                }
            ]
            ```

            This attribute can also be configured from the config with the
            `SECURITY_SCHEMES` configuration key. Defaults to `None`.
        spec_callback: It stores the function object registered by
            [`spec_processor`][apiflask.APIFlask.spec_processor]. You can also
            pass a callback function to it directly without using `spec_processor`.
            Example:

            ```python
            def update_spec(spec):
                spec['title'] = 'Updated Title'
                return spec

            app.spec_callback = update_spec
            ```

        error_callback: It stores the function object registered by
            [`error_processor`][apiflask.APIFlask.error_processor]. You can also
            pass a callback function to it directly without using `error_processor`.
            See the docstring of `error_processor` for more details.
            Example:

            ```python
            def my_error_handler(error):
                return {
                    'status_code': error.status_code,
                    'message': error.message,
                    'detail': error.detail
                }, error.status_code, error.headers

            app.error_processor = my_error_handler
            ```

        schema_name_resolver: It stores the function that used to decided the schema name.
            The schema name resolver should accept the schema object as argument and return
            the name.
            Example:

            ```python
            # this is the default schema name resolver used in APIFlask
            def schema_name_resolver(schema):
                name = schema.__class__.__name__  # get schema class name
                if schema.partial:  # add a "Update" suffix for partial schema
                    name += 'Update'
                return name

            app.schema_name_resolver = schema_name_resolver
            ```

    *Version changed: 1.0*

    - Add instance attribute `security_schemes` as an alias of config `SECURITY_SCHEMES`.

    *Version changed: 0.9.0*

    - Add instance attribute `schema_name_resolver`.
    """

    openapi_version: str = ConfigAttribute('OPENAPI_VERSION')  # type: ignore
    tags: TagsType | None = ConfigAttribute('TAGS')  # type: ignore
    servers: list[dict[str, str]] | None = ConfigAttribute('SERVERS')  # type: ignore
    info: dict[str, str | dict] | None = ConfigAttribute('INFO')  # type: ignore
    description: str | None = ConfigAttribute('DESCRIPTION')  # type: ignore
    contact: dict[str, str] | None = ConfigAttribute('CONTACT')  # type: ignore
    license: dict[str, str] | None = ConfigAttribute('LICENSE')  # type: ignore
    external_docs: dict[str, str] | None = ConfigAttribute('EXTERNAL_DOCS')  # type: ignore
    terms_of_service: str | None = ConfigAttribute('TERMS_OF_SERVICE')  # type: ignore
    security_schemes: dict[str, t.Any] | None = ConfigAttribute('SECURITY_SCHEMES')  # type: ignore

    def __init__(
        self,
        import_name: str,
        title: str = 'APIFlask',
        version: str = '0.1.0',
        spec_path: str | None = '/openapi.json',
        docs_path: str | None = '/docs',
        docs_oauth2_redirect_path: str | None = '/docs/oauth2-redirect',
        docs_oauth2_redirect_path_external: bool = False,
        docs_ui: str = 'swagger-ui',
        openapi_blueprint_url_prefix: str | None = None,
        json_errors: bool = True,
        enable_openapi: bool = True,
        spec_plugins: list[BasePlugin] | None = None,
        static_url_path: str | None = None,
        static_folder: str = 'static',
        static_host: str | None = None,
        host_matching: bool = False,
        subdomain_matching: bool = False,
        template_folder: str = 'templates',
        instance_path: str | None = None,
        instance_relative_config: bool = False,
        root_path: str | None = None,
    ) -> None:
        """Make an app instance.

        Arguments:
            import_name: The name of the application package, usually
                `__name__`. This helps locate the `root_path` for the
                application.
            title: The title of the API (openapi.info.title), defaults to "APIFlask".
                You can change it to the name of your API (e.g., "Pet API").
            version: The version of the API (openapi.info.version), defaults to "0.1.0".
            spec_path: The path to OpenAPI Spec documentation. It
                defaults to `/openapi.json`, if the path ends with `.yaml`
                or `.yml`, the YAML format of the OAS will be returned.
            docs_path: The path to API UI documentation, defaults to `/docs`.
            docs_ui: The UI of API documentation, one of `swagger-ui` (default), `redoc`,
                `elements`, `rapidoc`, and `rapipdf`.
            docs_oauth2_redirect_path: The path to Swagger UI OAuth redirect.
            docs_oauth2_redirect_path_external: If `True`, the `docs_oauth2_redirect_path`
                will be an external URL.
            openapi_blueprint_url_prefix: The url prefix of the OpenAPI blueprint. This
                prefix will append before all the OpenAPI-related paths (`sepc_path`,
                `docs_path`, etc.), defaults to `None`.
            json_errors: If `True`, APIFlask will return a JSON response for HTTP errors.
            enable_openapi: If `False`, will disable OpenAPI spec and API docs views.
            spec_plugins: List of apispec-compatible plugins (subclasses of `apispec.BasePlugin`),
                defaults to `None`. The `MarshmallowPlugin` for apispec is already included
                by default, so it doesn't need to be provided here.

        Other keyword arguments are directly passed to `flask.Flask`.

        *Version changed: 2.0.0*

        - Remove the deprecated `redoc_path` parameter.

        *Version changed: 1.2.0*

        - Add `spec_plugins` parameter.

        *Version changed: 1.1.0*

        - Add `docs_ui` parameter.

        *Version changed: 0.7.0*

        - Add `openapi_blueprint_url_prefix` parameter.
        """
        super().__init__(
            import_name,
            static_url_path=static_url_path,
            static_folder=static_folder,
            static_host=static_host,
            host_matching=host_matching,
            subdomain_matching=subdomain_matching,
            template_folder=template_folder,
            instance_path=instance_path,
            instance_relative_config=instance_relative_config,
            root_path=root_path,
        )

        # Set default config
        self.config.from_object('apiflask.settings')

        self.title = title
        self.version = version
        self.spec_path = spec_path
        self.docs_ui = docs_ui
        self.docs_path = docs_path
        self.docs_oauth2_redirect_path = docs_oauth2_redirect_path
        self.docs_oauth2_redirect_path_external = docs_oauth2_redirect_path_external
        self.openapi_blueprint_url_prefix = openapi_blueprint_url_prefix
        self.enable_openapi = enable_openapi
        self.json_errors = json_errors

        self.spec_callback: SpecCallbackType | None = None
        self.error_callback: ErrorCallbackType = self._error_handler
        self.schema_name_resolver = self._schema_name_resolver

        self.spec_plugins: list[BasePlugin] = spec_plugins or []
        self._spec: dict | str | None = None
        self._auth_blueprints: dict[str, t.Dict[str, t.Any]] = {}
        self._auths: set[HTTPAuthType | MultiAuth] = set()

        self._register_openapi_blueprint()
        self._register_error_handlers()

    def _register_error_handlers(self) -> None:
        """Register default error handlers for HTTPError and WerkzeugHTTPException.

        *Version changed: 0.9.0*

        - Always pass an `HTTPError` instance to error handlers.
        """

        @self.errorhandler(HTTPError)  # type: ignore
        def handle_http_errors(error: HTTPError) -> ResponseReturnValueType:
            return self.error_callback(error)

        if self.json_errors:
            self._apply_error_callback_to_werkzeug_errors()

    def _apply_error_callback_to_werkzeug_errors(self) -> None:
        @self.errorhandler(WerkzeugHTTPException)  # type: ignore
        def handle_werkzeug_errors(e: WerkzeugHTTPException) -> ResponseReturnValueType:
            headers = dict(e.get_headers())
            # remove the original MIME header
            del headers['Content-Type']
            error = HTTPError(e.code, message=e.name, headers=headers)
            return self.error_callback(error)

    # TODO: remove this function when we drop the Flask < 2.2 support
    # List return values are supported in Flask 2.2
    def make_response(self, rv) -> Response:
        """Patch the make_response form Flask to allow returning list as JSON.
        *Version added: 1.1.0*
        """
        if isinstance(rv, list):
            rv = jsonify(rv)
        elif isinstance(rv, tuple) and isinstance(rv[0], list):
            rv = (jsonify(rv[0]), *rv[1:])
        return super().make_response(rv)

    @staticmethod
    def _error_handler(error: HTTPError) -> ResponseReturnValueType:
        """The default error handler.

        Arguments:
            error: An instance of [`HTTPError`][apiflask.exceptions.HTTPError].

        *Version changed: 0.10.0*

        - Remove the `status_code` field from the response.
        - Add `HTTPError.extra_data` to the response body.
        """
        body = {'detail': error.detail, 'message': error.message, **error.extra_data}
        return body, error.status_code, error.headers

    def error_processor(self, f: ErrorCallbackType) -> ErrorCallbackType:
        """A decorator to register a custom error response processor function.

        The decorated callback function will be called in the following situations:

        - Any HTTP exception is raised by Flask when handling request.
        - A validation error happened when parsing a request.
        - An exception triggered with [`HTTPError`][apiflask.exceptions.HTTPError]
        - An exception triggered with [`abort`][apiflask.exceptions.abort].

        You can still register a specific error handler for a specific error code
        or exception with the `app.errorhandler(code_or_execution)` decorator,
        in that case, the return value of the specific error handler will be used as the
        response when the corresponding error or exception happened.

        The callback function must accept an error object as argument and return a valid
        response.

        Examples:

        ```python
        @app.error_processor
        def my_error_processor(error):
            return {
                'status_code': error.status_code,
                'message': error.message,
                'detail': error.detail,
                **error.extra_data
            }, error.status_code, error.headers
        ```

        The error object is an instance of [`HTTPError`][apiflask.exceptions.HTTPError],
        so you can get error information via it's attributes:

        - status_code: If the error is triggered by a validation error, the value will be
          422 (default) or the value you passed in config `VALIDATION_ERROR_STATUS_CODE`.
          If the error is triggered by [`HTTPError`][apiflask.exceptions.HTTPError]
          or [`abort`][apiflask.exceptions.abort], it will be the status code
          you passed. Otherwise, it will be the status code set by Werkzeug when
          processing the request.
        - message: The error description for this error, either you passed or grabbed from
          Werkzeug.
        - detail: The detail of the error. When the validation error happens, it will
          be filled automatically in the following structure:

            ```python
            "<location>": {
                "<field_name>": ["<error_message>", ...],
                "<field_name>": ["<error_message>", ...],
                ...
            },
            "<location>": {
                ...
            },
            ...
            ```

          The value of `location` can be `json` (i.e., request body) or `query`
          (i.e., query string) depending on the place where the validation error
          happened.
        - headers: The value will be `{}` unless you pass it in `HTTPError` or `abort`.
        - extra_data: Additional error information.

        If you want, you can rewrite the whole response body to anything you like:

        ```python
        @app.error_processor
        def my_error_processor(error):
            body = {'error_detail': error.detail, **error.extra_data}
            return body, error.status_code, error.headers
        ```

        However, I would recommend keeping the `detail` in the response since it contains
        the detailed information about the validation error when the validation error
        happened.

        *Version changed: 1.0*

        - Apply this error processor to normal HTTP errors even when
          `json_error` is set to `False` when creating `APIFlask` instance.

        *Version changed: 0.7.0*

        - Support registering an async callback function.
        """
        self.error_callback = self.ensure_sync(f)
        self._apply_error_callback_to_werkzeug_errors()
        return f

    def _register_openapi_blueprint(self) -> None:
        """Register a blueprint for OpenAPI support.

        The name of the blueprint is "openapi". This blueprint will hold the view
        functions for spec file and API docs.

        *Version changed: 2.3.3*

        - Add 'docs_oauth2_redirect_path_external' parameter to support absolute redirect url.

        *Version changed: 2.1.0*

        - Inject the OpenAPI endpoints decorators.

        *Version changed: 1.1.0*

        - Deprecate the redoc view at /redoc path.

        *Version changed: 0.7.0*

        - The format of the spec now rely on the `SPEC_FORMAT` config.
        """
        bp = Blueprint('openapi', __name__, url_prefix=self.openapi_blueprint_url_prefix)

        if self.spec_path:

            @bp.route(self.spec_path)
            @self._apply_decorators(config_name='SPEC_DECORATORS')
            def spec():
                if self.config['SPEC_FORMAT'] == 'json':
                    response = jsonify(self._get_spec('json'))
                    response.mimetype = self.config['JSON_SPEC_MIMETYPE']
                    return response
                return (
                    self._get_spec('yaml'),
                    200,
                    {'Content-Type': self.config['YAML_SPEC_MIMETYPE']},
                )

        if self.docs_path:
            if self.docs_ui not in ui_templates:
                valid_values = list(ui_templates.keys())
                raise ValueError(
                    f'Invalid docs_ui value, expected one of {valid_values!r}, '
                    f'got {self.docs_ui!r}.'
                )

            @bp.route(self.docs_path)
            @self._apply_decorators(config_name='DOCS_DECORATORS')
            def docs():
                docs_oauth2_redirect_path = None
                if self.docs_ui == 'swagger-ui':
                    if self.docs_oauth2_redirect_path_external:
                        docs_oauth2_redirect_path = url_for(
                            'openapi.swagger_ui_oauth_redirect', _external=True
                        )
                    else:
                        docs_oauth2_redirect_path = self.docs_oauth2_redirect_path

                return render_template_string(
                    ui_templates[self.docs_ui],
                    title=self.title,
                    version=self.version,
                    oauth2_redirect_path=docs_oauth2_redirect_path,
                )

            if self.docs_ui == 'swagger-ui':
                if self.docs_oauth2_redirect_path:

                    @bp.route(self.docs_oauth2_redirect_path)
                    @self._apply_decorators(config_name='SWAGGER_UI_OAUTH_REDIRECT_DECORATORS')
                    def swagger_ui_oauth_redirect() -> str:
                        return render_template_string(swagger_ui_oauth2_redirect_template)

        if self.enable_openapi and (self.spec_path or self.docs_path):
            self.register_blueprint(bp)

    def _get_spec(self, spec_format: str | None = None, force_update: bool = False) -> dict | str:
        """Get the current OAS document file.

        This method will return the cached spec on the first call. If you want
        to get the latest spec, set the `force_update` to `True` or use the
        public attribute `app.spec`, which will always return the newly generated
        spec when you call it.

        If the config `SYNC_LOCAL_SPEC` is `True`, the local spec
        specified in config `LOCAL_SPEC_PATH` will be automatically updated
        when the spec changes.

        Arguments:
            spec_format: The format of the spec file, one of `'json'`, `'yaml'`
                and `'yml'`, defaults to the `SPEC_FORMAT` config.
            force_update: If true, will generate the spec for every call instead
                of using the cache.

        *Version changed: 0.7.0*

        - The default format now rely on the `SPEC_FORMAT` config.
        - Support to sync local spec file.

        *Version changed: 0.7.1*

        - Rename the method name to `_get_spec`.
        - Add the `force_update` parameter.

        *Version changed: 1.3.0*

        - Add the `SPEC_PROCESSOR_PASS_OBJECT` config to control the argument type
          when calling the spec processor.
        """
        if spec_format is None:
            spec_format = self.config['SPEC_FORMAT']
        if self._spec is None or force_update:
            spec_object: APISpec = self._generate_spec()
            if self.spec_callback:
                if self.config['SPEC_PROCESSOR_PASS_OBJECT']:
                    self._spec = self.spec_callback(
                        spec_object  # type: ignore
                    ).to_dict()
                else:
                    self._spec = self.spec_callback(spec_object.to_dict())
            else:
                self._spec = spec_object.to_dict()
            if spec_format in ['yml', 'yaml']:
                from apispec.yaml_utils import dict_to_yaml

                self._spec = dict_to_yaml(self._spec)  # type: ignore
        # sync local spec
        if self.config['SYNC_LOCAL_SPEC']:
            spec_path = self.config['LOCAL_SPEC_PATH']
            if spec_path is None:
                raise TypeError('The spec path (LOCAL_SPEC_PATH) should be a valid path string.')
            spec: str
            if spec_format == 'json':
                spec = json.dumps(self._spec, indent=self.config['LOCAL_SPEC_JSON_INDENT'])
            else:
                spec = str(self._spec)
            with open(spec_path, 'w') as f:
                f.write(spec)
        return self._spec  # type: ignore

    def spec_processor(self, f: SpecCallbackType) -> SpecCallbackType:
        """A decorator to register a spec handler callback function.

        You can register a function to update the spec. The callback function
        should accept the spec as an argument and return it in the end. The
        callback function will be called when generating the spec file.

        Examples:

        ```python
        @app.spec_processor
        def update_spec(spec):
            spec['info']['title'] = 'Updated Title'
            return spec
        ```

        Notice the format of the spec is depends on the the value of configuration
        variable `SPEC_FORMAT` (defaults to `'json'`):

        - `'json'` -> dict
        - `'yaml'` -> string

        *Version Changed: 0.7.0*

        - Support registering an async callback function.
        """
        self.spec_callback = self.ensure_sync(f)
        return f

    @property
    def spec(self) -> dict | str:
        """Get the current OAS document file.

        This property will call `app._get_spec()` method and set the
        `force_update` parameter to `True`.

        *Version changed: 0.7.1*

        - Generate the spec on every call.
        """
        return self._get_spec(force_update=True)

    @staticmethod
    def _schema_name_resolver(schema: type[Schema]) -> str:
        """Default schema name resolver."""
        adapter = registry.create_adapter(schema)
        return adapter.get_schema_name()

    def _make_info(self) -> dict:
        """Make OpenAPI info object."""
        info: dict
        if self.info:
            info = self.info
        else:
            info = {}
        if self.contact:
            info['contact'] = self.contact
        if self.license:
            info['license'] = self.license
        if self.terms_of_service:
            info['termsOfService'] = self.terms_of_service
        if self.description:
            info['description'] = self.description
        return info

    def _make_tags(self) -> list[dict[str, t.Any]]:
        """Make OpenAPI tags object."""
        tags: TagsType | None = self.tags
        if tags is not None:
            # convert simple tags list into standard OpenAPI tags
            if isinstance(tags[0], str):
                for index, tag_name in enumerate(tags):
                    tags[index] = {'name': tag_name}  # type: ignore
        else:
            tags: list[dict[str, t.Any]] = []  # type: ignore
            if self.config['AUTO_TAGS']:
                # auto-generate tags from blueprints
                for blueprint_name, blueprint in self.blueprints.items():
                    if (
                        blueprint_name == 'openapi'
                        or not hasattr(blueprint, 'enable_openapi')
                        or not blueprint.enable_openapi
                    ):  # type: ignore
                        continue
                    tag: dict[str, t.Any] = get_tag(blueprint, blueprint_name)  # type: ignore
                    tags.append(tag)  # type: ignore
        return tags  # type: ignore

    def _collect_security_info(self) -> t.Tuple[list[str], list[HTTPAuthType]]:
        """Detect `auth_required` on blueprint before_request functions and view functions."""
        # security schemes
        auth_names: set[str] = set()

        def _register_base_auth(auth: HTTPAuthType) -> None:
            if not isinstance(auth, SecurityScheme):
                raise TypeError('Unknown authentication scheme.')

            if auth.name in auth_names:
                warnings.warn(
                    f"The auth scheme name '{auth.name}' has existed, "
                    'so it will be overwritten.',
                    stacklevel=2,
                )

            self._auths.add(auth)
            auth_names.add(auth.name)

        def _update_auth_info(auth: HTTPAuthType | MultiAuth) -> None:
            if isinstance(auth, MultiAuth):
                self._auths.add(auth)
                for base_auth in auth._auths:
                    _register_base_auth(base_auth)
                return

            _register_base_auth(auth)

        # collect auth info on blueprint before_request functions
        for blueprint_name, funcs in self.before_request_funcs.items():
            # skip app-level before_request functions (blueprint_name is None)
            if blueprint_name is None or not self.blueprints[blueprint_name].enable_openapi:  # type: ignore
                continue
            for f in funcs:
                if hasattr(f, '_spec'):  # pragma: no cover
                    auth = f._spec.get('auth')  # type: ignore
                    if auth is not None and auth not in self._auths:
                        self._auth_blueprints[blueprint_name] = {
                            'auth': auth,
                            'roles': f._spec.get('roles'),  # type: ignore
                        }
                        _update_auth_info(auth)
        # collect auth info on view functions
        for rule in self.url_map.iter_rules():
            view_func: ViewFuncType = self.view_functions[rule.endpoint]  # type: ignore
            if hasattr(view_func, '_spec'):
                auth = view_func._spec.get('auth')
                if auth is not None and auth not in self._auths:
                    _update_auth_info(auth)
            # method views
            if hasattr(view_func, '_method_spec'):
                for method_spec in view_func._method_spec.values():
                    auth = method_spec.get('auth')
                    if auth is not None and auth not in self._auths:
                        _update_auth_info(auth)

        # only base authentications have a schema, multiple authentication consists of base authentications # noqa: E501
        base_auths = [auth for auth in self._auths if isinstance(auth, _AuthBase)]
        return [security_schema.name for security_schema in base_auths], [
            security_schema for security_schema in base_auths
        ]

    def _generate_spec(self) -> APISpec:
        """Generate the spec, return an instance of `apispec.APISpec`.

        *Version changed: 1.3.0*

        - Support setting custom response content type.

        *Version changed: 1.2.1*

        - Set default `servers` value.

        *Version changed: 0.10.0*

        - Add support for `operationId`.
        - Add support for response `links`.

        *Version changed: 0.9.0*

        - Add base response customization support.

        *Version changed: 0.8.0*

        - Add automatic 404 response support.
        """
        # Track registered schema classes to avoid duplicates
        # Maps schema class id to registered name
        registered_schema_classes: dict[int, str] = {}

        kwargs: dict = {}
        if self.servers:
            kwargs['servers'] = self.servers
        else:
            if self.config['AUTO_SERVERS'] and has_request_context():
                kwargs['servers'] = [{'url': request.url_root}]
        if self.external_docs:
            kwargs['externalDocs'] = self.external_docs

        # Keep marshmallow plugin for backwards compatibility
        try:
            self._ma_plugin: MarshmallowPlugin = MarshmallowPlugin(
                schema_name_resolver=self.schema_name_resolver  # type: ignore
            )
            spec_plugins: list[BasePlugin] = [self._ma_plugin, *self.spec_plugins]
        except ImportError:
            # If marshmallow is not available, just use custom plugins
            self._ma_plugin = None  # type: ignore
            spec_plugins = self.spec_plugins

        spec: APISpec = APISpec(
            title=self.title,
            version=self.version,
            openapi_version=self.config['OPENAPI_VERSION'],
            plugins=spec_plugins,
            info=self._make_info(),
            tags=self._make_tags(),
            **kwargs,
        )

        # configure flask-marshmallow URL types if marshmallow plugin is available
        if self._ma_plugin is not None:
            # configure flask-marshmallow URL types
            self._ma_plugin.converter.field_mapping[fields.URLFor] = ('string', 'url')  # type: ignore
            self._ma_plugin.converter.field_mapping[fields.AbsoluteURLFor] = (  # type: ignore
                'string',
                'url',
            )
            if sqla is not None:  # pragma: no cover
                self._ma_plugin.converter.field_mapping[sqla.HyperlinkRelated] = (  # type: ignore
                    'string',
                    'url',
                )

        auth_names, auth_schemes = self._collect_security_info()
        security, security_schemes = get_security_and_security_schemes(auth_names, auth_schemes)

        if self.config['SECURITY_SCHEMES'] is not None:
            security_schemes.update(self.config['SECURITY_SCHEMES'])

        for name, scheme in security_schemes.items():
            spec.components.security_scheme(name, scheme)

        # paths
        paths: dict[str, dict[str, t.Any]] = {}
        rules: list[t.Any] = sorted(
            list(self.url_map.iter_rules()), key=lambda rule: len(rule.rule)
        )
        for rule in rules:
            operations: dict[str, t.Any] = {}
            view_func: ViewFuncType = self.view_functions[rule.endpoint]  # type: ignore
            # skip endpoints from openapi blueprint and the built-in static endpoint
            if rule.endpoint in default_bypassed_endpoints:
                continue
            blueprint_name: str | None = None  # type: ignore
            if '.' in rule.endpoint:
                blueprint_name: str = rule.endpoint.rsplit('.', 1)[0]  # type: ignore
                blueprint = self.blueprints.get(blueprint_name)  # type: ignore
                if blueprint is None:
                    # just a normal view with dots in its endpoint, reset blueprint_name
                    blueprint_name = None
                else:
                    if (
                        rule.endpoint == (f'{blueprint_name}.static')
                        or not hasattr(blueprint, 'enable_openapi')
                        or not blueprint.enable_openapi
                    ):  # type: ignore
                        continue
            # add a default 200 response for bare views
            if not hasattr(view_func, '_spec'):
                if not inspect.ismethod(view_func) and self.config['AUTO_200_RESPONSE']:  # type: ignore
                    view_func._spec = {'response': default_response}
                else:
                    continue  # pragma: no cover
            # method views
            if hasattr(view_func, '_method_spec'):
                skip = True
                for method, method_spec in view_func._method_spec.items():
                    if method_spec.get('no_spec'):
                        if self.config['AUTO_200_RESPONSE']:
                            view_func._method_spec[method]['response'] = default_response
                            skip = False
                    else:
                        skip = False
                if skip:
                    continue
            # skip views flagged with @app.doc(hide=True)
            if view_func._spec.get('hide'):
                continue

            # operation tags
            operation_tags: list[str] | None = None
            if view_func._spec.get('tags'):
                operation_tags = view_func._spec.get('tags')
            else:
                # use blueprint name as tag
                if self.tags is None and self.config['AUTO_TAGS'] and blueprint_name is not None:
                    blueprint = self.blueprints[blueprint_name]
                    operation_tags = get_operation_tags(blueprint, blueprint_name)  # type: ignore

            for method in ['GET', 'POST', 'PUT', 'PATCH', 'DELETE']:
                if method not in rule.methods:
                    continue
                # method views
                if hasattr(view_func, '_method_spec'):
                    if method not in view_func._method_spec:
                        continue  # pragma: no cover
                    view_func._spec = view_func._method_spec[method]

                    if view_func._spec.get('no_spec') and not self.config['AUTO_200_RESPONSE']:
                        continue
                    if (
                        view_func._spec.get('generated_summary')
                        and not self.config['AUTO_OPERATION_SUMMARY']
                    ):
                        view_func._spec['summary'] = ''
                    if (
                        view_func._spec.get('generated_description')
                        and not self.config['AUTO_OPERATION_DESCRIPTION']
                    ):
                        view_func._spec['description'] = ''
                    if view_func._spec.get('hide'):
                        continue
                    if view_func._spec.get('tags'):
                        operation_tags = view_func._spec.get('tags')
                    else:
                        if (
                            self.tags is None
                            and self.config['AUTO_TAGS']
                            and blueprint_name is not None
                        ):
                            blueprint = self.blueprints[blueprint_name]
                            operation_tags = get_operation_tags(blueprint, blueprint_name)  # type: ignore

                # operation parameters
                # Process parameters using schema adapters to handle both marshmallow and Pydantic
                parameters = []
                for schema, location in view_func._spec.get('args', []):
                    try:
                        # Use schema_to_parameters for proper handling of different schema types
                        schema_params = openapi_helper.schema_to_parameters(
                            schema, location=location
                        )
                        parameters.extend(schema_params)
                    except Exception:
                        # Fallback to original behavior for unknown schema types
                        parameters.append({'in': location, 'schema': schema})

                operation: dict[str, t.Any] = {
                    'parameters': parameters,
                    'responses': {},
                }
                if operation_tags:
                    operation['tags'] = operation_tags

                # summary
                if view_func._spec.get('summary'):
                    operation['summary'] = view_func._spec.get('summary')
                else:
                    # auto-generate summary from dotstring or view function name
                    if self.config['AUTO_OPERATION_SUMMARY']:
                        operation['summary'] = get_path_summary(view_func)  # type: ignore

                # description
                if view_func._spec.get('description'):
                    operation['description'] = view_func._spec.get('description')
                else:
                    # auto-generate description from dotstring
                    if self.config['AUTO_OPERATION_DESCRIPTION']:
                        docs = [
                            line.strip() for line in (view_func.__doc__ or '').strip().split('\n')
                        ]
                        if len(docs) > 1:
                            # use the remain lines of docstring as description
                            operation['description'] = '\n'.join(docs[1:]).strip()

                # deprecated
                if view_func._spec.get('deprecated'):
                    operation['deprecated'] = view_func._spec.get('deprecated')

                # operationId
                operation_id = view_func._spec.get('operation_id')
                if operation_id is None:
                    if self.config['AUTO_OPERATION_ID']:
                        operation['operationId'] = (
                            f"{method.lower()}_{rule.endpoint.replace('.', '_')}"
                        )
                else:
                    operation['operationId'] = operation_id

                # responses
                if view_func._spec.get('response'):
                    schema = view_func._spec.get('response')['schema']
                    # Check if the adapter was created with many=True (for list types)
                    schema_adapter_many = view_func._spec.get('response').get(
                        'schema_adapter_many', False
                    )

                    status_code: str = str(view_func._spec.get('response')['status_code'])
                    description: str = (
                        view_func._spec.get('response')['description']
                        or self.config['SUCCESS_DESCRIPTION']
                    )
                    example = view_func._spec.get('response')['example']
                    examples = view_func._spec.get('response')['examples']
                    links = view_func._spec.get('response')['links']
                    content_type = view_func._spec.get('response')['content_type']
                    headers = view_func._spec.get('response')['headers']
                    self._add_response(
                        spec,
                        registered_schema_classes,
                        operation,
                        status_code,
                        schema,
                        description,
                        example=example,
                        examples=examples,
                        links=links,
                        content_type=content_type,
                        headers_schema=headers,
                        schema_adapter_many=schema_adapter_many,
                    )
                else:
                    # add a default 200 response for views without using @app.output
                    # or @app.doc(responses={...})
                    if not view_func._spec.get('responses') and self.config['AUTO_200_RESPONSE']:
                        self._add_response(
                            spec,
                            registered_schema_classes,
                            operation,
                            '200',
                            {},
                            self.config['SUCCESS_DESCRIPTION'],
                        )

                # add validation error response
                if self.config['AUTO_VALIDATION_ERROR_RESPONSE'] and (
                    view_func._spec.get('body') or view_func._spec.get('args')
                ):
                    status_code: str = str(  # type: ignore
                        self.config['VALIDATION_ERROR_STATUS_CODE']
                    )
                    description: str = self.config[  # type: ignore
                        'VALIDATION_ERROR_DESCRIPTION'
                    ]
                    schema: SchemaType = self.config['VALIDATION_ERROR_SCHEMA']  # type: ignore
                    self._add_response_with_schema(
                        spec,
                        registered_schema_classes,
                        operation,
                        status_code,
                        schema,
                        'ValidationError',
                        description,
                    )

                # add authentication error response
                has_bp_level_auth = (
                    blueprint_name is not None and blueprint_name in self._auth_blueprints
                )
                view_func_auth = view_func._spec.get('auth')
                custom_security = view_func._spec.get('security')
                operation_extensions = view_func._spec.get('extensions')
                if self.config['AUTO_AUTH_ERROR_RESPONSE'] and (
                    has_bp_level_auth or view_func_auth or custom_security
                ):
                    status_code: str = str(  # type: ignore
                        self.config['AUTH_ERROR_STATUS_CODE']
                    )
                    description: str = self.config['AUTH_ERROR_DESCRIPTION']  # type: ignore
                    schema: SchemaType = self.config['HTTP_ERROR_SCHEMA']  # type: ignore
                    self._add_response_with_schema(
                        spec,
                        registered_schema_classes,
                        operation,
                        status_code,
                        schema,
                        'HTTPError',
                        description,
                    )

                # add 404 error response
                if self.config['AUTO_404_RESPONSE'] and rule.arguments:
                    description: str = self.config['NOT_FOUND_DESCRIPTION']  # type: ignore
                    schema: SchemaType = self.config['HTTP_ERROR_SCHEMA']  # type: ignore
                    self._add_response_with_schema(
                        spec,
                        registered_schema_classes,
                        operation,
                        '404',
                        schema,
                        'HTTPError',
                        description,
                    )

                if view_func._spec.get('responses'):
                    responses: ResponsesType = view_func._spec.get('responses')
                    # turn status_code list to dict {status_code: reason_phrase}
                    if isinstance(responses, list):
                        responses: dict[int, str] = {}  # type: ignore
                        for status_code in view_func._spec.get('responses'):
                            responses[  # type: ignore
                                status_code
                            ] = get_reason_phrase(int(status_code), '')
                    for status_code, value in responses.items():  # type: ignore
                        status_code: str = str(status_code)  # type: ignore
                        # custom complete response spec
                        if isinstance(value, dict):
                            existing_response = operation['responses'].setdefault(status_code, {})
                            existing_response_content = existing_response.setdefault('content', {})
                            existing_response_content.update(value.get('content', {}))
                            if (new_description := value.get('description')) is not None:
                                existing_response['description'] = new_description
                            continue
                        else:
                            description = value
                        # overwrite existing response description
                        if status_code in operation['responses']:
                            if not isinstance(
                                view_func._spec.get('responses'), list
                            ):  # pragma: no cover
                                operation['responses'][status_code]['description'] = description
                            continue
                        # add error response schema for error responses
                        if status_code.startswith('4') or status_code.startswith('5'):
                            schema: SchemaType = self.config['HTTP_ERROR_SCHEMA']  # type: ignore
                            self._add_response_with_schema(
                                spec,
                                registered_schema_classes,
                                operation,
                                status_code,
                                schema,
                                'HTTPError',
                                description,
                            )
                        else:  # add default response for other responses
                            self._add_response(
                                spec,
                                registered_schema_classes,
                                operation,
                                status_code,
                                {},
                                description,
                            )

                # requestBody
                if view_func._spec.get('body'):
                    content_types = view_func._spec.get('content_type')
                    if not isinstance(content_types, list):
                        content_types = [content_types]
                    operation['requestBody'] = {'content': {}}
                    for content_type in content_types:
                        body_schema_obj = view_func._spec['body']

                        # Handle body schema registration and referencing
                        try:
                            # Check if this is a schema object (marshmallow/pydantic schema)
                            # Plain dicts have __class__.__name__ == 'dict'
                            is_schema_obj = (
                                hasattr(body_schema_obj, '__class__')
                                and hasattr(body_schema_obj.__class__, '__name__')
                                and body_schema_obj.__class__.__name__ != 'dict'
                            )
                            # Skip if schema is already a reference (dict with $ref key)
                            if isinstance(body_schema_obj, dict) and '$ref' in body_schema_obj:
                                body_schema = body_schema_obj
                            elif is_schema_obj:
                                # Register schema and get reference
                                body_schema = self._register_schema_and_get_ref(
                                    spec, registered_schema_classes, body_schema_obj
                                )
                            else:
                                # Fallback to inline schema for plain dicts or other types
                                body_schema = body_schema_obj
                        except Exception:
                            # If anything fails, fall back to passing schema as-is
                            body_schema = body_schema_obj

                        operation['requestBody']['content'][content_type] = {
                            'schema': body_schema,
                        }
                        if view_func._spec.get('body_example'):
                            example = view_func._spec.get('body_example')
                            operation['requestBody']['content'][content_type]['example'] = example
                        if view_func._spec.get('body_examples'):
                            examples = view_func._spec.get('body_examples')
                            operation['requestBody']['content'][content_type]['examples'] = examples

                # security
                if custom_security:  # custom security
                    # TODO: validate the security name and the format
                    operation['security'] = []
                    operation_security = custom_security
                    if isinstance(operation_security, str):  # 'A' -> [{'A': []}]
                        operation['security'] = [{operation_security: []}]
                    elif isinstance(operation_security, list):
                        # ['A', 'B'] -> [{'A': []}, {'B': []}]
                        if isinstance(operation_security[0], str):
                            operation['security'] = [{name: []} for name in operation_security]
                        else:
                            operation['security'] = operation_security
                    else:
                        raise ValueError('The operation security must be a string or a list.')
                else:
                    if has_bp_level_auth:
                        bp_auth_info = self._auth_blueprints[blueprint_name]  # type: ignore
                        operation['security'] = [
                            {security[bp_auth_info['auth']]: bp_auth_info['roles']}
                        ]

                    # view-wide auth
                    if view_func_auth:
                        if isinstance(view_func_auth, MultiAuth):
                            operation['security'] = [
                                {base_auth.name: view_func._spec['roles']}
                                for base_auth in view_func_auth._auths
                            ]
                        else:
                            operation['security'] = [
                                {security[view_func_auth]: view_func._spec['roles']}
                            ]

                operations[method.lower()] = operation

                if operation_extensions:
                    for extension, value in operation_extensions.items():
                        operation[extension] = value

            # parameters
            path_arguments: t.Iterable = re.findall(r'<(([^<:]+:)?([^>]+))>', rule.rule)
            if path_arguments and not (
                hasattr(view_func, '_spec')
                and view_func._spec.get('omit_default_path_parameters', False)
            ):
                arguments: list[dict[str, str]] = []
                for _, argument_type, argument_name in path_arguments:
                    argument = get_argument(argument_type, argument_name)
                    arguments.append(argument)

                for _, operation in operations.items():
                    operation['parameters'] = arguments + operation['parameters']

            path: str = re.sub(r'<([^<:]+:)?', '{', rule.rule).replace('>', '}')
            if path not in paths:
                paths[path] = operations
            else:
                paths[path].update(operations)

        for path, operations in paths.items():
            # sort by method before adding them to the spec
            sorted_operations: dict[str, t.Any] = {}
            for method in ['get', 'post', 'put', 'patch', 'delete']:
                if method in operations:
                    sorted_operations[method] = operations[method]
            spec.path(path=path, operations=sorted_operations)

        return spec

    def _apply_decorators(self, config_name: str):
        """Apply the decorators to the OpenAPI endpoints at runtime.

        Arguments:
            config_name: The config name to get the list of decorators.
        """

        def decorator(f):
            @wraps(f)
            def wrapper(*args, **kwargs):
                decorated_func = f
                decorators = self.config[config_name]
                if decorators:
                    for decorator in decorators:
                        decorated_func = decorator(decorated_func)
                return decorated_func(*args, **kwargs)

            return wrapper

        return decorator

    def _register_schema_and_get_ref(
        self,
        spec: APISpec,
        registered_schema_classes: dict[int, str],
        schema_obj: t.Any,
    ) -> dict[str, str]:
        """Register a schema and return its reference.

        Arguments:
            spec: The APISpec object
            registered_schema_classes: Dictionary tracking registered schema classes
            schema_obj: The schema object to register

        Returns:
            A dictionary with $ref key pointing to the registered schema
        """
        # For classes (like Pydantic models), use the class itself
        # For instances (like Marshmallow schemas), use the class
        if isinstance(schema_obj, type):
            schema_class_id = id(schema_obj)
        else:
            schema_class_id = id(schema_obj.__class__)

        if schema_class_id in registered_schema_classes and not (
            hasattr(schema_obj, 'partial') and schema_obj.partial
        ):  # type: ignore
            # Reuse the name from the first registration
            schema_name = registered_schema_classes[schema_class_id]
        else:
            # Get the schema name for registration
            schema_name = self.schema_name_resolver(schema_obj)

            # Handle name conflicts with different schema classes
            if schema_name in spec.components.schemas:
                schema_name = get_unique_schema_name(spec, schema_name)

            # Register schema - convert to OpenAPI dict for non-marshmallow schemas
            try:
                # Try to detect if it's a marshmallow schema
                adapter = registry.create_adapter(schema_obj)
                if adapter.schema_type == 'marshmallow':
                    # For marshmallow, pass the schema object to MarshmallowPlugin
                    spec.components.schema(schema_name, schema=schema_obj)
                else:
                    # For other schema types (Pydantic, etc), convert to dict first
                    schema_dict = openapi_helper.schema_to_json_schema(schema_obj)

                    # Extract and register nested $defs (for Pydantic)
                    nested_defs = extract_pydantic_defs(schema_dict, schema_name)
                    for nested_name, nested_schema in nested_defs.items():
                        if nested_name not in spec.components.schemas:
                            spec.components.schema(nested_name, nested_schema)

                    # Register the main schema
                    spec.components.schema(schema_name, schema_dict)
            except Exception:
                # Fallback: try to register as-is
                spec.components.schema(schema_name, schema=schema_obj)

            # Track this schema class
            registered_schema_classes[schema_class_id] = schema_name

        # Return reference
        return {'$ref': f'#/components/schemas/{schema_name}'}

    def _add_response(
        self,
        spec: APISpec,
        registered_schema_classes: dict[int, str],
        operation: dict,
        status_code: str,
        schema: SchemaType | dict,
        description: str,
        example: t.Any | None = None,
        examples: dict[str, t.Any] | None = None,
        links: dict[str, t.Any] | None = None,
        content_type: str | None = 'application/json',
        headers_schema: SchemaType | None = None,
        schema_adapter_many: bool = False,
    ) -> None:
        """Add response to operation.

        *Version changed: 2.1.0*

        - Add parameter `headers_schema`.

        *Version changed: 1.3.0*

        - Add parameter `content_type`.

        *Version changed: 0.10.0*

        - Add `links` parameter.
        """
        # Register schema if it's a schema object (not FileSchema, EmptySchema, or plain dict)
        if status_code != '204' and schema:
            is_schema_obj = (
                hasattr(schema, '__class__')
                and hasattr(schema.__class__, '__name__')
                and schema.__class__.__name__ != 'dict'
                and not isinstance(schema, (FileSchema, EmptySchema))
                and not (isinstance(schema, dict) and '$ref' in schema)
            )

            if is_schema_obj:
                # Register schema and get reference
                schema = self._register_schema_and_get_ref(spec, registered_schema_classes, schema)

            # Wrap schema in array type if many=True
            if schema_adapter_many and isinstance(schema, dict) and '$ref' in schema:
                schema = {'type': 'array', 'items': schema}

        base_schema: OpenAPISchemaType | None = self.config['BASE_RESPONSE_SCHEMA']
        data_key: str = self.config['BASE_RESPONSE_DATA_KEY']
        if base_schema is not None:
            base_schema_spec: dict[str, t.Any]
            if isinstance(base_schema, type):
                # Convert schema class to instance, then to full JSON schema
                # Use schema_to_json_schema to get complete schema with properties
                base_schema_spec = openapi_helper.schema_to_json_schema(base_schema)
            elif isinstance(base_schema, dict):
                base_schema_spec = base_schema
            else:
                raise TypeError(_bad_schema_message)
            if data_key not in base_schema_spec['properties']:  # type: ignore
                raise RuntimeError(
                    f'The data key {data_key!r} is not found in the base response schema spec.'
                )
            base_schema_spec['properties'][data_key] = schema  # type: ignore
            schema = base_schema_spec

        operation['responses'][status_code] = {}
        if status_code != '204':
            if isinstance(schema, FileSchema):
                schema = {'type': schema.type, 'format': schema.format}
            elif isinstance(schema, EmptySchema):
                schema = {}
            operation['responses'][status_code]['content'] = {content_type: {'schema': schema}}
        operation['responses'][status_code]['description'] = description
        if example is not None:
            operation['responses'][status_code]['content'][content_type]['example'] = example
        if examples is not None:
            operation['responses'][status_code]['content'][content_type]['examples'] = examples
        if links is not None:
            operation['responses'][status_code]['links'] = links
        if headers_schema is not None:
            # Use openapi_helper to convert headers schema to parameters
            header_params = openapi_helper.schema_to_parameters(headers_schema, location='headers')
            headers = {header['name']: header for header in header_params}
            for header in headers.values():
                header.pop('in', None)
                header.pop('name', None)
            operation['responses'][status_code]['headers'] = headers

    def _add_response_with_schema(
        self,
        spec: APISpec,
        registered_schema_classes: dict[int, str],
        operation: dict,
        status_code: str,
        schema: OpenAPISchemaType,
        schema_name: str,
        description: str,
    ) -> None:
        """Add response with given schema to operation."""
        if isinstance(schema, type):
            schema = schema()
            self._add_response(
                spec, registered_schema_classes, operation, status_code, schema, description
            )
        elif isinstance(schema, dict):
            if schema_name not in spec.components.schemas:
                spec.components.schema(schema_name, schema)
            schema_ref = {'$ref': f'#/components/schemas/{schema_name}'}
            self._add_response(
                spec, registered_schema_classes, operation, status_code, schema_ref, description
            )
        else:
            raise TypeError(_bad_schema_message)



import openapi_spec_validator as osv
import pytest

from apiflask import APIKeyCookieAuth
from apiflask import APIKeyHeaderAuth
from apiflask import APIKeyQueryAuth
from apiflask import HTTPBasicAuth
from apiflask import HTTPTokenAuth
from apiflask.security import MultiAuth


def test_httpbasicauth_security_scheme(app, client):
    auth = HTTPBasicAuth()

    @app.get('/')
    @app.auth_required(auth)
    def foo():
        pass

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert 'BasicAuth' in rv.json['components']['securitySchemes']
    assert rv.json['components']['securitySchemes']['BasicAuth'] == {
        'type': 'http',
        'scheme': 'basic',
    }


def test_httptokenauth_security_scheme(app, client):
    auth = HTTPTokenAuth()

    @app.get('/')
    @app.auth_required(auth)
    def foo():
        pass

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert 'BearerAuth' in rv.json['components']['securitySchemes']
    assert rv.json['components']['securitySchemes']['BearerAuth'] == {
        'scheme': 'bearer',
        'type': 'http',
    }


def test_deprecated_apikey_auth_security_scheme(app, client):
    auth = HTTPTokenAuth('apiKey', header='X-API-Key')

    @app.get('/')
    @app.auth_required(auth)
    def foo():
        pass

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert 'ApiKeyAuth' in rv.json['components']['securitySchemes']
    assert rv.json['components']['securitySchemes']['ApiKeyAuth'] == {
        'type': 'apiKey',
        'name': 'X-API-Key',
        'in': 'header',
    }


def test_apikey_header_auth_security_scheme(app, client):
    auth = APIKeyHeaderAuth()

    @app.get('/')
    @app.auth_required(auth)
    def foo():
        pass

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert 'ApiKeyAuth' in rv.json['components']['securitySchemes']
    assert rv.json['components']['securitySchemes']['ApiKeyAuth'] == {
        'type': 'apiKey',
        'name': 'X-API-Key',
        'in': 'header',
    }


def test_apikey_cookie_auth_security_scheme(app, client):
    auth = APIKeyCookieAuth()

    @app.get('/')
    @app.auth_required(auth)
    def foo():
        pass

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert 'ApiKeyAuth' in rv.json['components']['securitySchemes']
    assert rv.json['components']['securitySchemes']['ApiKeyAuth'] == {
        'type': 'apiKey',
        'name': 'X-API-Key',
        'in': 'cookie',
    }


def test_apikey_query_auth_security_scheme(app, client):
    auth = APIKeyQueryAuth()

    @app.get('/')
    @app.auth_required(auth)
    def foo():
        pass

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert 'ApiKeyAuth' in rv.json['components']['securitySchemes']
    assert rv.json['components']['securitySchemes']['ApiKeyAuth'] == {
        'type': 'apiKey',
        'name': 'X-API-Key',
        'in': 'query',
    }


def test_custom_security_scheme_name(app, client):
    basic_auth = HTTPBasicAuth(security_scheme_name='basic_auth')
    apikey_header_auth = APIKeyHeaderAuth(security_scheme_name='myAPIKey')

    @app.get('/foo')
    @app.auth_required(basic_auth)
    def foo():
        pass

    @app.get('/bar')
    @app.auth_required(apikey_header_auth)
    def bar():
        pass

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert 'basic_auth' in rv.json['components']['securitySchemes']
    assert 'myAPIKey' in rv.json['components']['securitySchemes']
    assert rv.json['components']['securitySchemes']['basic_auth'] == {
        'type': 'http',
        'scheme': 'basic',
    }
    assert rv.json['components']['securitySchemes']['myAPIKey'] == {
        'type': 'apiKey',
        'name': 'X-API-Key',
        'in': 'header',
    }
    assert 'basic_auth' in rv.json['paths']['/foo']['get']['security'][0]
    assert 'myAPIKey' in rv.json['paths']['/bar']['get']['security'][0]


def test_unknown_auth_security_scheme(app):
    from flask_httpauth import HTTPDigestAuth

    auth = HTTPDigestAuth()

    @app.get('/')
    @app.auth_required(auth)
    def foo():
        pass

    with pytest.raises(TypeError):
        app.spec


def test_multiple_auth_names(app, client):
    auth1 = HTTPBasicAuth()
    auth2 = HTTPBasicAuth()
    auth3 = HTTPBasicAuth()

    @app.get('/foo')
    @app.auth_required(auth1)
    def foo():
        pass

    @app.get('/bar')
    @app.auth_required(auth2)
    def bar():
        pass

    @app.get('/baz')
    @app.auth_required(auth3)
    def baz():
        pass

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert 'BasicAuth' in rv.json['components']['securitySchemes']
    assert len(rv.json['components']['securitySchemes']) == 1


def test_security_schemes_description(app, client):
    basic_auth = HTTPBasicAuth(description='some description for basic auth')
    token_auth = HTTPTokenAuth(description='some description for bearer auth')
    apikey_header_auth = APIKeyHeaderAuth(description='some description for apikey auth')

    @app.get('/foo')
    @app.auth_required(basic_auth)
    def foo():
        pass

    @app.get('/bar')
    @app.auth_required(token_auth)
    def bar():
        pass

    @app.get('/baz')
    @app.auth_required(apikey_header_auth)
    def baz():
        pass

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert 'BasicAuth' in rv.json['components']['securitySchemes']
    assert 'BearerAuth' in rv.json['components']['securitySchemes']
    assert 'ApiKeyAuth' in rv.json['components']['securitySchemes']
    assert rv.json['components']['securitySchemes']['BasicAuth'] == {
        'type': 'http',
        'scheme': 'basic',
        'description': 'some description for basic auth',
    }
    assert rv.json['components']['securitySchemes']['BearerAuth'] == {
        'type': 'http',
        'scheme': 'bearer',
        'description': 'some description for bearer auth',
    }
    assert rv.json['components']['securitySchemes']['ApiKeyAuth'] == {
        'type': 'apiKey',
        'name': 'X-API-Key',
        'in': 'header',
        'description': 'some description for apikey auth',
    }


def test_multi_auth(app, client):
    basic_auth = HTTPBasicAuth()
    token_auth = HTTPTokenAuth()
    multi_auth = MultiAuth(basic_auth, token_auth)

    @app.route('/foo')
    @app.auth_required(multi_auth)
    def foo():
        pass

    rv = client.get('/foo')
    assert rv.status_code == 401
    assert rv.headers['Content-Type'] == 'application/json'
    assert 'message' in rv.json
    assert rv.json['message'] == 'Unauthorized'
    assert 'WWW-Authenticate' in rv.headers



import openapi_spec_validator as osv

from apiflask import Schema
from apiflask.fields import Field
from apiflask.fields import Integer
from apiflask.fields import String


class PetIn(Schema):
    """Input schema for pet."""

    name = String(required=True)
    age = Integer()


class PetOut(Schema):
    """Output schema for pet."""

    id = Integer()
    name = String()
    age = Integer()


class BaseResponseSchema(Schema):
    """Base response wrapper schema."""

    message = String()
    status = Integer()
    data = Field()


def test_schema_name_resolver_with_adapter(app, client):
    """Test that schema name resolver uses adapter."""

    @app.post('/pets')
    @app.input(PetIn)
    @app.output(PetOut, status_code=201)
    def create_pet(json_data):
        return {'id': 1, 'name': json_data['name'], 'age': json_data.get('age', 0)}

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)

    # Check that schema names are resolved correctly
    assert 'PetIn' in rv.json['components']['schemas']
    assert 'PetOut' in rv.json['components']['schemas']


def test_body_schema_registration_with_reference(app, client):
    """Test that body schemas are registered and referenced correctly."""

    @app.post('/items')
    @app.input(PetIn)
    def create_item(json_data):
        return {'success': True}

    rv = client.get('/openapi.json')
    assert rv.status_code == 200

    # Validate OpenAPI spec
    osv.validate(rv.json)

    # Check that schema is registered in components
    assert 'PetIn' in rv.json['components']['schemas']

    # Check that request body uses $ref
    request_body = rv.json['paths']['/items']['post']['requestBody']
    schema = request_body['content']['application/json']['schema']
    assert '$ref' in schema
    assert schema['$ref'] == '#/components/schemas/PetIn'


def test_body_schema_registration_duplicate_handling(app, client):
    """Test that duplicate schemas are handled correctly."""

    @app.post('/pets')
    @app.input(PetIn)
    def create_pet(json_data):
        return {'id': 1}

    @app.put('/pets/<int:pet_id>')
    @app.input(PetIn)
    def update_pet(pet_id, json_data):
        return {'id': pet_id}

    rv = client.get('/openapi.json')
    assert rv.status_code == 200

    # Validate OpenAPI spec
    osv.validate(rv.json)

    # Both endpoints should reference the same schema
    post_schema = rv.json['paths']['/pets']['post']['requestBody']['content']['application/json'][
        'schema'
    ]
    put_schema = rv.json['paths']['/pets/{pet_id}']['put']['requestBody']['content'][
        'application/json'
    ]['schema']

    assert post_schema == put_schema
    assert post_schema['$ref'] == '#/components/schemas/PetIn'

    # Schema should only be registered once
    assert 'PetIn' in rv.json['components']['schemas']
    assert 'PetIn1' not in rv.json['components']['schemas']


def test_body_schema_registration_with_dict(app, client):
    """Test that dict schemas are handled correctly."""
    dict_schema = {'name': String(required=True), 'value': Integer()}

    @app.post('/data')
    @app.input(dict_schema, schema_name='DataInput')
    def create_data(json_data):
        return {'success': True}

    rv = client.get('/openapi.json')
    assert rv.status_code == 200

    # Validate OpenAPI spec
    osv.validate(rv.json)

    # Check that schema is registered
    assert 'DataInput' in rv.json['components']['schemas']

    # Check reference
    request_body = rv.json['paths']['/data']['post']['requestBody']
    schema = request_body['content']['application/json']['schema']
    assert schema['$ref'] == '#/components/schemas/DataInput'


def test_body_schema_registration_generated_schemas(app, client):
    """Test that generated schemas get unique names."""
    dict_schema1 = {'field1': String()}
    dict_schema2 = {'field2': Integer()}

    @app.post('/endpoint1')
    @app.input(dict_schema1)
    def endpoint1(json_data):
        return {}

    @app.post('/endpoint2')
    @app.input(dict_schema2)
    def endpoint2(json_data):
        return {}

    rv = client.get('/openapi.json')
    assert rv.status_code == 200

    # Validate OpenAPI spec
    osv.validate(rv.json)

    # Generated schemas should have unique names
    schemas = rv.json['components']['schemas']
    generated_schemas = [name for name in schemas if name.startswith('Generated')]

    assert len(generated_schemas) == 2
    assert 'GeneratedSchema' in generated_schemas
    assert 'GeneratedSchema1' in generated_schemas


def test_parameters_use_openapi_helper(app, client):
    """Test that query parameters use openapi_helper."""

    class QuerySchema(Schema):
        search = String()
        limit = Integer()

    @app.get('/search')
    @app.input(QuerySchema, location='query')
    def search(query_data):
        return {'results': []}

    rv = client.get('/openapi.json')
    assert rv.status_code == 200

    # Validate OpenAPI spec
    osv.validate(rv.json)

    # Check that parameters are extracted correctly
    params = rv.json['paths']['/search']['get']['parameters']
    assert len(params) == 2

    param_names = [p['name'] for p in params]
    assert 'search' in param_names
    assert 'limit' in param_names

    # All should be query parameters
    assert all(p['in'] == 'query' for p in params)


def test_headers_schema_uses_openapi_helper(app, client):
    """Test that response headers use openapi_helper."""

    class HeaderSchema(Schema):
        x_rate_limit = Integer(data_key='X-RateLimit-Limit')
        x_rate_remaining = Integer(data_key='X-RateLimit-Remaining')

    @app.get('/status')
    @app.output(PetOut, headers=HeaderSchema)
    def get_status():
        return {'id': 1, 'name': 'Test', 'age': 5}

    rv = client.get('/openapi.json')
    assert rv.status_code == 200

    # Validate OpenAPI spec
    osv.validate(rv.json)

    # Check response headers
    response = rv.json['paths']['/status']['get']['responses']['200']
    assert 'headers' in response

    headers = response['headers']
    assert 'X-RateLimit-Limit' in headers
    assert 'X-RateLimit-Remaining' in headers

    # Headers should have schema definitions
    assert 'schema' in headers['X-RateLimit-Limit']
    assert headers['X-RateLimit-Limit']['schema']['type'] == 'integer'


def test_base_response_schema_uses_openapi_helper(app, client):
    """Test that base response schema uses openapi_helper."""
    app.config['BASE_RESPONSE_SCHEMA'] = BaseResponseSchema
    app.config['BASE_RESPONSE_DATA_KEY'] = 'data'

    @app.get('/items')
    @app.output(PetOut)
    def get_items():
        return {'message': 'Success', 'status': 200, 'data': {'id': 1, 'name': 'Test', 'age': 5}}

    rv = client.get('/openapi.json')
    assert rv.status_code == 200

    # Validate OpenAPI spec
    osv.validate(rv.json)

    # Check that response schema includes base response structure
    response = rv.json['paths']['/items']['get']['responses']['200']
    schema = response['content']['application/json']['schema']

    assert schema['type'] == 'object'
    assert 'properties' in schema
    assert 'message' in schema['properties']
    assert 'status' in schema['properties']
    assert 'data' in schema['properties']


def test_base_response_schema_with_dict(app, client):
    """Test base response schema with dict configuration."""
    app.config['BASE_RESPONSE_SCHEMA'] = {
        'properties': {
            'success': {'type': 'boolean'},
            'data': {'type': 'object'},
        },
        'type': 'object',
    }
    app.config['BASE_RESPONSE_DATA_KEY'] = 'data'

    @app.get('/test')
    @app.output(PetOut)
    def test_endpoint():
        return {'success': True, 'data': {'id': 1, 'name': 'Test', 'age': 5}}

    rv = client.get('/openapi.json')
    assert rv.status_code == 200

    # Validate OpenAPI spec
    osv.validate(rv.json)

    response = rv.json['paths']['/test']['get']['responses']['200']
    schema = response['content']['application/json']['schema']

    assert 'properties' in schema
    assert 'success' in schema['properties']
    assert 'data' in schema['properties']


def test_path_parameters_use_openapi_helper(app, client):
    """Test that path parameters are processed correctly."""

    class PathSchema(Schema):
        pet_id = Integer(required=True)

    @app.get('/pets/<int:pet_id>')
    @app.input(PathSchema, location='path')
    def get_pet(pet_id, path_data):
        return {'id': pet_id, 'name': 'Test', 'age': 5}

    rv = client.get('/openapi.json')
    assert rv.status_code == 200

    # Validate OpenAPI spec
    osv.validate(rv.json)

    # Check path parameters
    params = rv.json['paths']['/pets/{pet_id}']['get']['parameters']
    path_params = [p for p in params if p['in'] == 'path']

    assert len(path_params) >= 1
    param_names = [p['name'] for p in path_params]
    assert 'pet_id' in param_names


def test_multiple_content_types_body_schema(app, client):
    """Test body schema registration with multiple content types."""

    @app.post('/data')
    @app.input(PetIn, location='json_or_form')
    def create_data(data):
        return {'success': True}

    rv = client.get('/openapi.json')
    assert rv.status_code == 200

    # Validate OpenAPI spec
    osv.validate(rv.json)

    request_body = rv.json['paths']['/data']['post']['requestBody']

    # Should have both content types
    assert 'application/json' in request_body['content']
    assert 'application/x-www-form-urlencoded' in request_body['content']

    # Both should reference the same schema
    json_schema = request_body['content']['application/json']['schema']
    form_schema = request_body['content']['application/x-www-form-urlencoded']['schema']

    assert json_schema == form_schema
    assert '$ref' in json_schema


def test_unique_schema_names_for_different_schemas(app, client):
    """Test that different schemas with same name get unique references."""

    class ItemSchema(Schema):
        name = String()

    # Create two different schemas with potential name collision
    @app.post('/items1')
    @app.input(ItemSchema)
    def create_item1(json_data):
        return {}

    @app.post('/items2')
    @app.input(ItemSchema)
    def create_item2(json_data):
        return {}

    rv = client.get('/openapi.json')
    assert rv.status_code == 200

    # Validate OpenAPI spec
    osv.validate(rv.json)

    # Same schema class should be registered once and reused
    assert 'ItemSchema' in rv.json['components']['schemas']

    post1_schema = rv.json['paths']['/items1']['post']['requestBody']['content'][
        'application/json'
    ]['schema']
    post2_schema = rv.json['paths']['/items2']['post']['requestBody']['content'][
        'application/json'
    ]['schema']

    # Should reference the same schema
    assert post1_schema == post2_schema



import openapi_spec_validator as osv

from .schemas import Foo
from .schemas import Header
from .schemas import Pagination
from .schemas import Query


def test_spec_path_summary_description_from_docs(app, client):
    @app.route('/users')
    @app.output(Foo)
    def get_users():
        """Get Users"""
        pass

    @app.route('/users/<id>', methods=['PUT'])
    @app.output(Foo)
    def update_user(id):
        """Update User

        Update a user with specified ID.
        """
        pass

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert rv.json['paths']['/users']['get']['summary'] == 'Get Users'
    assert rv.json['paths']['/users/{id}']['put']['summary'] == 'Update User'
    assert (
        rv.json['paths']['/users/{id}']['put']['description'] == 'Update a user with specified ID.'
    )


def test_spec_path_parameters_registration(app, client):
    @app.route('/strings/<some_string>')
    @app.output(Foo)
    def get_string(some_string):
        pass

    @app.route('/floats/<float:some_float>', methods=['POST'])
    @app.output(Foo)
    def get_float(some_float):
        pass

    @app.route('/integers/<int:some_integer>', methods=['PUT'])
    @app.output(Foo)
    def get_integer(some_integer):
        pass

    @app.route('/users/<int:user_id>/articles/<int:article_id>')
    @app.output(Foo)
    def get_article(user_id, article_id):
        pass

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert rv.json['paths']['/strings/{some_string}']['get']['parameters'][0]['in'] == 'path'
    assert (
        rv.json['paths']['/strings/{some_string}']['get']['parameters'][0]['name'] == 'some_string'
    )
    assert (
        rv.json['paths']['/strings/{some_string}']['get']['parameters'][0]['schema']['type']
        == 'string'
    )
    assert (
        rv.json['paths']['/floats/{some_float}']['post']['parameters'][0]['schema']['type']
        == 'number'
    )
    assert (
        rv.json['paths']['/integers/{some_integer}']['put']['parameters'][0]['schema']['type']
        == 'integer'
    )
    assert (
        rv.json['paths']['/users/{user_id}/articles/{article_id}']['get']['parameters'][0]['name']
        == 'user_id'
    )
    assert (
        rv.json['paths']['/users/{user_id}/articles/{article_id}']['get']['parameters'][1]['name']
        == 'article_id'
    )


def test_spec_path_summary_auto_generation(app, client):
    @app.route('/users')
    @app.output(Foo)
    def get_users():
        pass

    @app.route('/users/<id>', methods=['PUT'])
    @app.output(Foo)
    def update_user(id):
        pass

    @app.route('/users/<id>', methods=['DELETE'])
    @app.output(Foo)
    def delete_user(id):
        """Summary from Docs

        Delete a user with specified ID.
        """
        pass

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert rv.json['paths']['/users']['get']['summary'] == 'Get Users'
    assert rv.json['paths']['/users/{id}']['put']['summary'] == 'Update User'
    assert rv.json['paths']['/users/{id}']['delete']['summary'] == 'Summary from Docs'
    assert (
        rv.json['paths']['/users/{id}']['delete']['description']
        == 'Delete a user with specified ID.'
    )


def test_path_arguments_detection(app, client):
    @app.route('/foo/<bar>')
    @app.output(Foo)
    def pattern1(bar):
        pass

    @app.route('/<foo>/bar')
    @app.output(Foo)
    def pattern2(foo):
        pass

    @app.route('/<int:foo>/<bar>/baz')
    @app.output(Foo)
    def pattern3(foo, bar):
        pass

    @app.route('/foo/<int:bar>/<int:baz>')
    @app.output(Foo)
    def pattern4(bar, baz):
        pass

    @app.route('/<int:foo>/<bar>/<float:baz>')
    @app.output(Foo)
    def pattern5(foo, bar, baz):
        pass

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert '/foo/{bar}' in rv.json['paths']
    assert '/{foo}/bar' in rv.json['paths']
    assert '/{foo}/{bar}/baz' in rv.json['paths']
    assert '/foo/{bar}/{baz}' in rv.json['paths']
    assert '/{foo}/{bar}/{baz}' in rv.json['paths']
    assert (
        rv.json['paths']['/{foo}/{bar}/{baz}']['get']['parameters'][0]['schema']['type']
        == 'integer'
    )
    assert (
        rv.json['paths']['/{foo}/{bar}/{baz}']['get']['parameters'][1]['schema']['type'] == 'string'
    )
    assert (
        rv.json['paths']['/{foo}/{bar}/{baz}']['get']['parameters'][2]['schema']['type'] == 'number'
    )


def test_path_arguments_order(app, client):
    @app.route('/<foo>/bar')
    @app.input(Query, location='query')
    @app.output(Foo)
    def path_and_query(foo, query_data):
        pass

    @app.route('/<foo>/<bar>')
    @app.output(Foo)
    def two_path_variables(foo, bar):
        pass

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert '/{foo}/bar' in rv.json['paths']
    assert '/{foo}/{bar}' in rv.json['paths']
    assert rv.json['paths']['/{foo}/bar']['get']['parameters'][0]['name'] == 'foo'
    assert rv.json['paths']['/{foo}/bar']['get']['parameters'][1]['name'] == 'id'
    assert rv.json['paths']['/{foo}/{bar}']['get']['parameters'][0]['name'] == 'foo'
    assert rv.json['paths']['/{foo}/{bar}']['get']['parameters'][1]['name'] == 'bar'


def test_parameters_registration(app, client):
    @app.route('/foo')
    @app.input(Query, location='query')
    @app.output(Foo)
    def foo(query_data):
        pass

    @app.route('/bar')
    @app.input(Query, location='query', arg_name='query')
    @app.input(Pagination, location='query', arg_name='pagination')
    @app.input(Header, location='headers')
    def bar(query, pagination, headers_data):
        return {'query': query['id'], 'pagination': pagination, 'foo': headers_data['foo']}

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert '/foo' in rv.json['paths']
    assert '/bar' in rv.json['paths']
    assert rv.json['paths']['/foo']['get']['parameters'][0]['name'] == 'id'
    assert len(rv.json['paths']['/foo']['get']['parameters']) == 1
    assert len(rv.json['paths']['/bar']['get']['parameters']) == 4
    rv = client.get('/bar')
    assert rv.status_code == 200
    assert rv.json['query'] == 1
    assert rv.json['pagination']['page'] == 1
    assert rv.json['pagination']['per_page'] == 10
    assert rv.json['foo'] == 'bar'


def test_register_validation_error_response(app, client):
    error_code = str(app.config['VALIDATION_ERROR_STATUS_CODE'])

    @app.post('/foo')
    @app.input(Foo)
    def foo():
        pass

    @app.get('/bar')
    @app.input(Foo, location='query')
    def bar():
        pass

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert rv.json['paths']['/foo']['post']['responses'][error_code] is not None
    assert (
        rv.json['paths']['/foo']['post']['responses'][error_code]['description']
        == 'Validation error'
    )
    assert rv.json['paths']['/bar']['get']['responses'][error_code] is not None
    assert (
        rv.json['paths']['/bar']['get']['responses'][error_code]['description']
        == 'Validation error'
    )


def test_auto_404_error(app, client):
    @app.get('/foo')
    def foo():
        pass

    @app.get('/bar/<int:id>')
    def bar():
        pass

    @app.get('/baz/<int:id>')
    @app.doc(responses={404: 'Pet not found'})
    def baz():
        pass

    rv = client.get('/openapi.json')
    assert rv.status_code == 200
    osv.validate(rv.json)
    assert '404' not in rv.json['paths']['/foo']['get']['responses']
    assert '404' in rv.json['paths']['/bar/{id}']['get']['responses']
    assert rv.json['paths']['/bar/{id}']['get']['responses']['404']['description'] == 'Not found'
    assert (
        rv.json['paths']['/baz/{id}']['get']['responses']['404']['description'] == 'Pet not found'
    )



from __future__ import annotations

import typing as t

from .types import HTTPAuthType

if t.TYPE_CHECKING:  # pragma: no cover
    from .blueprint import APIBlueprint


default_bypassed_endpoints: list[str] = [
    'static',
    'openapi.spec',
    'openapi.docs',
    'openapi.redoc',
    'openapi.swagger_ui_oauth_redirect',
    '_debug_toolbar.static',  # Flask-DebugToolbar
]

default_response = {
    'schema': {},
    'status_code': 200,
    'description': None,
    'example': None,
    'examples': None,
    'links': None,
    'content_type': 'application/json',
    'headers': None,
}


def get_tag(blueprint: APIBlueprint, blueprint_name: str) -> dict[str, t.Any]:
    """Get tag from blueprint object."""
    tag: dict[str, t.Any]
    if blueprint.tag is not None:
        if isinstance(blueprint.tag, dict):
            tag = blueprint.tag
        else:
            tag = {'name': blueprint.tag}
    else:
        tag = {'name': blueprint_name.title()}
    return tag


def get_operation_tags(blueprint: APIBlueprint, blueprint_name: str) -> list[str]:
    """Get operation tag from blueprint object."""
    tags: list[str]
    if blueprint.tag is not None:
        if isinstance(blueprint.tag, dict):
            tags = [blueprint.tag['name']]
        else:
            tags = [blueprint.tag]
    else:
        tags = [blueprint_name.title()]
    return tags


def get_security_and_security_schemes(
    auth_names: list[str], auth_schemes: list[HTTPAuthType]
) -> tuple[dict[HTTPAuthType, str], dict[str, dict[str, str]]]:
    """Make security and security schemes from given auth names and schemes."""
    security: dict[HTTPAuthType, str] = {}
    security_schemes: dict[str, dict[str, str]] = {}
    for name, auth in zip(auth_names, auth_schemes):  # noqa: B905
        security[auth] = name
        security_schemes[name] = auth.get_security_scheme()
    return security, security_schemes


def get_path_summary(func: t.Callable, fallback: str | None = None) -> str:
    """Get path summary from the name or docstring of the view function."""
    summary: str
    docs: list = (func.__doc__ or '').strip().split('\n')
    if docs[0]:
        # Use the first line of docstring
        summary = docs[0]
    else:
        # Use the function name
        summary = fallback or ' '.join(func.__name__.split('_')).title()
    return summary


def get_path_description(func: t.Callable) -> str:
    """Get path description from the docstring of the view function."""
    docs = (func.__doc__ or '').strip().split('\n')
    if len(docs) > 1:
        # use the remain lines of docstring as description
        return '\n'.join(docs[1:]).strip()
    return ''


def get_argument(argument_type: str, argument_name: str) -> dict[str, t.Any]:
    """Make argument from given type and name."""
    argument: dict[str, t.Any] = {
        'in': 'path',
        'name': argument_name,
    }
    if argument_type == 'int:':
        argument['schema'] = {'type': 'integer'}
    elif argument_type == 'float:':
        argument['schema'] = {'type': 'number'}
    else:
        argument['schema'] = {'type': 'string'}
    return argument



import typing as t

from flask import current_app
from apiflask import APIFlask, HTTPBasicAuth, HTTPTokenAuth, MultiAuth, Schema, abort
from werkzeug.security import generate_password_hash, check_password_hash
from apiflask.fields import String
from authlib.jose import jwt, JoseError

app = APIFlask(__name__)
basic_auth = HTTPBasicAuth()
token_auth = HTTPTokenAuth()
multi_auth = MultiAuth(basic_auth, token_auth)
app.config['SECRET_KEY'] = 'secret-key'


class User:
    def __init__(self, id: int, username: str, password: str):
        self.id = id
        self.username = username
        self.password = generate_password_hash(password)

    def get_token(self):
        header = {'alg': 'HS256'}
        payload = {'id': self.id}
        return jwt.encode(header, payload, current_app.config['SECRET_KEY']).decode()


users = [
    User(1, 'lorem', 'foo'),
    User(2, 'ipsum', 'bar'),
]

username_map = {user.username: user for user in users}


def get_user_by_id(id: int) -> t.Union[User, None]:
    return tuple(filter(lambda u: u.id == id, users))[0]


@basic_auth.verify_password
def verify_password(username: str, password: str) -> t.Union[User, None]:
    if username in username_map and check_password_hash(username_map[username].password, password):
        return username_map[username]
    return None


@token_auth.verify_token
def verify_token(token: str) -> t.Union[User, None]:
    try:
        data = jwt.decode(
            token.encode('ascii'),
            current_app.config['SECRET_KEY'],
        )
        id = data['id']
        user = get_user_by_id(id)
    except JoseError:
        return None
    except IndexError:
        return None
    return user


class Token(Schema):
    token = String()


@app.post('/token/<int:id>')
@app.output(Token)
def get_token(id: int):
    if get_user_by_id(id) is None:
        abort(404)
    return {'token': f'Bearer {get_user_by_id(id).get_token()}'}


@app.route('/')
@app.auth_required(multi_auth)
def index():
    return f'Hello, {multi_auth.current_user.username}'



import typing as t

from apiflask import APIFlask, HTTPBasicAuth
from werkzeug.security import generate_password_hash, check_password_hash

app = APIFlask(__name__)
auth = HTTPBasicAuth()

users = {
    'userA': generate_password_hash('foo'),
    'userB': generate_password_hash('bar'),
}


# if you want to apply an authentication or other decorator
# to the endpoints used to serve the OpenAPI spec
# as well as the documentation UI, use the
# `SPEC_DECORATORS`, `DOCS_DECORATORS` and
# `SWAGGER_UI_OAUTH_REDIRECT_DECORATORS` configuration options
app.config['SPEC_DECORATORS'] = [app.auth_required(auth)]
app.config['DOCS_DECORATORS'] = [app.auth_required(auth)]


@auth.verify_password
def verify_password(username: str, password: str) -> t.Union[str, None]:
    if username in users and check_password_hash(users[username], password):
        return username
    return None


@app.route('/')
def index():
    return {
        'spec': '/openapi.json',
        'docs': '/docs',
    }



from __future__ import annotations

import typing as t
from functools import wraps

from flask import current_app
from flask import jsonify
from flask import request as flask_request
from flask import Response

from .helpers import _sentinel
from .schema_adapters import registry
from .schemas import FileSchema
from .types import DecoratedType
from .types import HTTPAuthType
from .types import OpenAPISchemaType
from .types import ResponseReturnValueType
from .types import ResponsesType
from .types import SchemaType
from .views import MethodView

if t.TYPE_CHECKING:
    from flask.sansio.scaffold import T_route  # noqa: F401

BODY_LOCATIONS = ['json', 'files', 'form', 'form_and_files', 'json_or_form']


def _annotate(f: t.Any, **kwargs: t.Any) -> None:
    if not hasattr(f, '_spec'):
        f._spec = {}
    for key, value in kwargs.items():
        f._spec[key] = value


def _ensure_sync(f):
    if hasattr(f, '_sync_ensured'):
        return f

    @wraps(f)
    def wrapper(*args, **kwargs):
        return current_app.ensure_sync(f)(*args, **kwargs)

    wrapper._sync_ensured = True
    return wrapper


class APIScaffold:
    """A base class for [`APIFlask`][apiflask.app.APIFlask] and
    [`APIBlueprint`][apiflask.blueprint.APIBlueprint].

    This class contains the route shortcut decorators (i.e. `get`, `post`, etc.) and
    API-related decorators (i.e. `auth_required`, `input`, `output`, `doc`).

    *Version added: 1.0*
    """

    def _method_route(
        self, method: str, rule: str, options: t.Any
    ) -> t.Callable[[T_route], T_route]:
        if 'methods' in options:
            raise RuntimeError('Use the "route" decorator to use the "methods" argument.')

        def decorator(f):
            if isinstance(f, type(MethodView)):
                raise RuntimeError(
                    'The route shortcuts cannot be used with "MethodView" classes, '
                    'use the "route" decorator instead.'
                )
            return self.route(rule, methods=[method], **options)(f)

        return decorator

    def get(self, rule: str, **options: t.Any) -> t.Callable[[T_route], T_route]:
        """Shortcut for `app.route()` or `app.route(methods=['GET'])`."""
        return self._method_route('GET', rule, options)

    def post(self, rule: str, **options: t.Any) -> t.Callable[[T_route], T_route]:
        """Shortcut for `app.route(methods=['POST'])`."""
        return self._method_route('POST', rule, options)

    def put(self, rule: str, **options: t.Any) -> t.Callable[[T_route], T_route]:
        """Shortcut for `app.route(methods=['PUT'])`."""
        return self._method_route('PUT', rule, options)

    def patch(self, rule: str, **options: t.Any) -> t.Callable[[T_route], T_route]:
        """Shortcut for `app.route(methods=['PATCH'])`."""
        return self._method_route('PATCH', rule, options)

    def delete(self, rule: str, **options: t.Any) -> t.Callable[[T_route], T_route]:
        """Shortcut for `app.route(methods=['DELETE'])`."""
        return self._method_route('DELETE', rule, options)

    def auth_required(
        self, auth: HTTPAuthType, roles: list | None = None, optional: str | None = None
    ) -> t.Callable[[DecoratedType], DecoratedType]:
        """Protect a view with provided authentication settings.

        > Be sure to put it under the routes decorators (i.e., `app.route`, `app.get`,
        `app.post`, etc.).

        Examples:

        ```python
        from apiflask import APIFlask, HTTPTokenAuth

        app = APIFlask(__name__)
        auth = HTTPTokenAuth()

        @app.get('/')
        @app.auth_required(auth)
        def hello():
            return 'Hello'!
        ```

        Arguments:
            auth: The `auth` object, an instance of
                [`HTTPBasicAuth`][apiflask.security.HTTPBasicAuth]
                or [`HTTPTokenAuth`][apiflask.security.HTTPTokenAuth]
                or [`HTTPAPIKeyAuth`][apiflask.security.HTTPAPIKeyAuth].
            roles: The selected roles to allow to visit this view, accepts a list of role names.
                See [Flask-HTTPAuth's documentation][_role]{target:_blank} for more details.
                [_role]: https://flask-httpauth.readthedocs.io/en/latest/#user-roles
            optional: Set to `True` to allow the view to execute even the authentication
                information is not included with the request, in which case the attribute
                `auth.current_user` will be `None`.

        *Version changed: 2.0.0*

        - Remove the deprecated `role` parameter.

        *Version changed: 1.0.0*

        - The `role` parameter is deprecated.

        *Version changed: 0.12.0*

        - Move to `APIFlask` and `APIBlueprint` classes.

        *Version changed: 0.4.0*

        - Add parameter `roles`.
        """

        def decorator(f):
            f = _ensure_sync(f)
            _annotate(f, auth=auth, roles=roles or [])
            return auth.login_required(role=roles, optional=optional)(f)

        return decorator

    def input(
        self,
        schema: SchemaType,
        location: str = 'json',
        arg_name: str | None = None,
        schema_name: str | None = None,
        example: t.Any | None = None,
        examples: dict[str, t.Any] | None = None,
        validation: bool = True,
        **kwargs: t.Any,
    ) -> t.Callable[[DecoratedType], DecoratedType]:
        """Add input settings for view functions.

        If the validation passed, the data will be injected into the view
        function as a keyword argument in the form of `dict` and named `{location}_data`.
        Otherwise, an error response with the detail of the validation result will be
        returned.

        > Be sure to put it under the routes decorators (i.e., `app.route`, `app.get`,
        `app.post`, etc.).

        Examples:

        ```python
        from apiflask import APIFlask

        app = APIFlask(__name__)

        @app.get('/')
        @app.input(PetIn, location='json')
        def hello(json_data):
            print(json_data)
            return 'Hello'!
        ```

        Arguments:
            schema: The marshmallow schema of the input data.
            location: The location of the input data, one of `'json'` (default),
                `'files'`, `'form'`, `'cookies'`, `'headers'`, `'query'`
                (same as `'querystring'`).
            arg_name: The name of the argument passed to the view function,
                defaults to `{location}_data`.
            schema_name: The schema name for dict schema, only needed when you pass
                a marshmallow schema dict (e.g., `{'name': String(required=True)}`) for `json`
                location.
            example: The example data in dict for request body, you should use either
                `example` or `examples`, not both.
            examples: Multiple examples for request body, you should pass a dict
                that contains multiple examples. Example:

                ```python
                {
                    'example foo': {  # example name
                        'summary': 'an example of foo',  # summary field is optional
                        'value': {'name': 'foo', 'id': 1}  # example value
                    },
                    'example bar': {
                        'summary': 'an example of bar',
                        'value': {'name': 'bar', 'id': 2}
                    },
                }
                ```
            validation: Flag to allow disabling of validation on input. Default to `True`.

        *Version changed: 2.2.2

        - Add parameter `validation` to allow disabling of validation on input.

        *Version changed: 2.0.0*

        - Always pass parsed data to view function as a keyword argument.
          The argument name will be in the form of `{location}_data`.

        *Version changed: 1.0*

        - Ensure only one input body location was used.
        - Add `form_and_files` and `json_or_form` (from webargs) location.
        - Rewrite `files` to act as `form_and_files`.
        - Use correct request content type for `form` and `files`.

        *Version changed: 0.12.0*

        - Move to APIFlask and APIBlueprint classes.

        *Version changed: 0.4.0*

        - Add parameter `examples`.
        """

        def decorator(f):
            f = _ensure_sync(f)

            is_body_location = location in BODY_LOCATIONS
            if is_body_location and hasattr(f, '_spec') and 'body' in f._spec:
                raise RuntimeError(
                    'When using the app.input() decorator, you can only declare one request '
                    'body location (one of "json", "form", "files", "form_and_files", '
                    'and "json_or_form").'
                )

            # Create schema adapter and use the instantiated schema for spec annotation
            # This ensures the marshmallow plugin receives schema instances
            adapter = registry.create_adapter(schema, schema_name=schema_name)
            annotation_schema = adapter.schema

            if location == 'json':
                _annotate(
                    f,
                    body=annotation_schema,
                    body_example=example,
                    body_examples=examples,
                    content_type='application/json',
                )
            elif location == 'form':
                _annotate(
                    f,
                    body=annotation_schema,
                    body_example=example,
                    body_examples=examples,
                    content_type='application/x-www-form-urlencoded',
                )
            elif location in ['files', 'form_and_files']:
                _annotate(
                    f,
                    body=annotation_schema,
                    body_example=example,
                    body_examples=examples,
                    content_type='multipart/form-data',
                )
            elif location == 'json_or_form':
                _annotate(
                    f,
                    body=annotation_schema,
                    body_example=example,
                    body_examples=examples,
                    content_type=['application/x-www-form-urlencoded', 'application/json'],
                )
            else:
                if not hasattr(f, '_spec') or f._spec.get('args') is None:
                    _annotate(f, args=[])
                if location in ['path', 'view_args']:
                    _annotate(f, omit_default_path_parameters=True)
                # TODO: Support set example for request parameters
                f._spec['args'].append((annotation_schema, location))

            arg_name_val = arg_name or f'{location}_data'

            # For marshmallow schemas, use the original webargs approach for compatibility
            if adapter.schema_type == 'marshmallow':
                from .schema_adapters.marshmallow import parser

                if not validation:

                    @wraps(f)
                    def wrapper(*args: t.Any, **kwargs: t.Any):
                        location_data = parser.load_location_data(
                            schema=annotation_schema, req=flask_request, location=location
                        )
                        kwargs[arg_name_val] = location_data
                        return f(*args, **kwargs)

                    return wrapper

                return parser.use_args(
                    annotation_schema, location=location, arg_name=arg_name_val, **kwargs
                )(f)

            # For other schema types (Pydantic, etc.), use the adapter system
            else:

                @wraps(f)
                def wrapper(*args: t.Any, **kwargs: t.Any):
                    location_data = adapter.validate_input(flask_request, location, **kwargs)
                    kwargs[arg_name_val] = location_data
                    return f(*args, **kwargs)

                return wrapper

        return decorator

    def output(
        self,
        schema: SchemaType,
        status_code: int = 200,
        description: str | None = None,
        schema_name: str | None = None,
        example: t.Any | None = None,
        examples: dict[str, t.Any] | None = None,
        links: dict[str, t.Any] | None = None,
        content_type: str | None = 'application/json',
        headers: SchemaType | None = None,
    ) -> t.Callable[[DecoratedType], DecoratedType]:
        """Add output settings for view functions.

        > Be sure to put it under the routes decorators (i.e., `app.route`, `app.get`,
        `app.post`, etc.).

        The decorator will format the return value of your view function with
        provided marshmallow schema. You can return a dict or an object (such
        as a model class instance of ORMs). APIFlask will handle the formatting
        and turn your return value into a JSON response.

        P.S. The output data will not be validated; it's a design choice of marshmallow.
        marshmallow 4.0 may be support the output validation.

        Examples:

        ```python
        from apiflask import APIFlask

        app = APIFlask(__name__)

        @app.get('/')
        @app.output(PetOut)
        def hello():
            return the_dict_or_object_match_petout_schema
        ```

        Arguments:
            schema: The schemas of the output data.
            status_code: The status code of the response, defaults to `200`.
            description: The description of the response.
            schema_name: The schema name for dict schema, only needed when you pass
                a schema dict (e.g., `{'name': String()}`).
            example: The example data in dict for response body, you should use either
                `example` or `examples`, not both.
            examples: Multiple examples for response body, you should pass a dict
                that contains multiple examples. Example:

                ```python
                {
                    'example foo': {  # example name
                        'summary': 'an example of foo',  # summary field is optional
                        'value': {'name': 'foo', 'id': 1}  # example value
                    },
                    'example bar': {
                        'summary': 'an example of bar',
                        'value': {'name': 'bar', 'id': 2}
                    },
                }
                ```
            links: The `links` of response. It accepts a dict which maps a link name to
                a link object. Example:

                ```python
                {
                    'getAddressByUserId': {
                        'operationId': 'getUserAddress',
                        'parameters': {
                            'userId': '$request.path.id'
                        }
                    }
                }
                ```

                See the [docs](https://apiflask.com/openapi/#response-links) for more details
                about setting response links.

            content_type: The content/media type of the response. It defaults to `application/json`.
            headers: The schemas of the headers.

        *Version changed: 2.1.0*

        - Add parameter `headers`.

        *Version changed: 2.0.0*

        - Don't change the status code to 204 for EmptySchema.

        *Version changed: 1.3.0*

        - Add parameter `content_type`.

        *Version changed: 0.12.0*

        - Move to APIFlask and APIBlueprint classes.

        *Version changed: 0.10.0*

        - Add `links` parameter.

        *Version changed: 0.9.0*

        - Add base response customization support.

        *Version changed: 0.6.0*

        - Support decorating async views.

        *Version changed: 0.5.2*

        - Return the `Response` object directly.

        *Version changed: 0.4.0*

        - Add parameter `examples`.
        """
        body_schema_adapter = registry.create_adapter(schema, schema_name=schema_name)
        body_schema = body_schema_adapter.schema

        headers_schema = None
        if headers is not None:
            headers_schema_adapter = registry.create_adapter(headers, schema_name=None)
            headers_schema = headers_schema_adapter.schema

        def decorator(f):
            f = _ensure_sync(f)
            _annotate(
                f,
                response={
                    'schema': body_schema,
                    'schema_adapter_many': body_schema_adapter.many,  # Store the many flag
                    'status_code': status_code,
                    'description': description,
                    'example': example,
                    'examples': examples,
                    'links': links,
                    'content_type': content_type,
                    'headers': headers_schema,
                },
            )

            def _jsonify(
                obj: t.Any,
                many: bool = _sentinel,  # type: ignore
                *args: t.Any,
                **kwargs: t.Any,
            ) -> Response:  # pragma: no cover
                """Serialize output using schema adapters."""
                if isinstance(body_schema, FileSchema):
                    return obj  # type: ignore
                # Handle many parameter
                if many is _sentinel:
                    # Check adapter's many flag first (for list[Model] syntax)
                    # Then check schema's many attribute (for marshmallow compatibility)
                    many = getattr(body_schema_adapter, 'many', False) or getattr(
                        schema, 'many', False
                    )  # type: ignore

                base_schema: OpenAPISchemaType = current_app.config['BASE_RESPONSE_SCHEMA']
                if base_schema is not None and status_code != 204:
                    data_key: str = current_app.config['BASE_RESPONSE_DATA_KEY']

                    if isinstance(obj, dict):
                        if data_key not in obj:
                            raise RuntimeError(
                                f'The data key {data_key!r} is not found in the returned dict.'
                            )
                        # Serialize the data part
                        obj[data_key] = body_schema_adapter.serialize_output(
                            obj[data_key], many=many
                        )
                    else:
                        if not hasattr(obj, data_key):
                            raise RuntimeError(
                                f'The data key {data_key!r} is not found in the returned object.'
                            )
                        # Serialize the data part
                        data_value = getattr(obj, data_key)
                        serialized_data = body_schema_adapter.serialize_output(
                            data_value, many=many
                        )
                        setattr(obj, data_key, serialized_data)

                    base_schema_adapter = registry.create_adapter(base_schema)
                    data = base_schema_adapter.serialize_output(obj)  # type: ignore
                else:
                    data = body_schema_adapter.serialize_output(obj, many=many)  # type: ignore
                return jsonify(data, *args, **kwargs)

            @wraps(f)
            def _response(*args: t.Any, **kwargs: t.Any) -> ResponseReturnValueType:
                rv = f(*args, **kwargs)
                if isinstance(rv, Response):
                    return rv
                if not isinstance(rv, tuple):
                    return _jsonify(rv), status_code
                json = _jsonify(rv[0])
                if len(rv) == 2:
                    rv = (json, rv[1]) if isinstance(rv[1], int) else (json, status_code, rv[1])
                elif len(rv) >= 3:
                    rv = (json, rv[1], rv[2])
                else:
                    rv = (json, status_code)
                return rv  # type: ignore

            return _response

        return decorator

    def doc(
        self,
        summary: str | None = None,
        description: str | None = None,
        tags: list[str] | None = None,
        responses: ResponsesType | None = None,
        deprecated: bool | None = None,
        hide: bool | None = None,
        operation_id: str | None = None,
        security: str | list[str | dict[str, list]] | None = None,
        extensions: dict[str, t.Any] | None = None,
    ) -> t.Callable[[DecoratedType], DecoratedType]:
        """Set up the OpenAPI Spec for view functions.

        > Be sure to put it under the routes decorators (i.e., `app.route`, `app.get`,
        `app.post`, etc.).

        Examples:

        ```python
        from apiflask import APIFlask

        app = APIFlask(__name__)

        @app.get('/')
        @app.doc(summary='Say hello', tags=['Foo'])
        def hello():
            return 'Hello'
        ```

        Arguments:
            summary: The summary of this endpoint. If not set, the name of the view function
                will be used. If your view function is named with `get_pet`, then the summary
                will be "Get Pet". If the view function has a docstring, then the first
                line of the docstring will be used. The precedence will be:

                ```
                @app.doc(summary='blah') > the first line of docstring > the view function name
                ```

            description: The description of this endpoint. If not set, the lines after the empty
                line of the docstring will be used.
            tags: A list of tag names of this endpoint, map the tags you passed in the `app.tags`
                attribute. If `app.tags` is not set, the blueprint name will be used as tag name.
            responses: The other responses for this view function, accepts a list of status codes
                (`[404, 418]`) or a dict in a format of either `{404: 'Not Found'}` or
                `{404: {'description': 'Not Found', 'content': {'application/json':
                {'schema': FooSchema}}}}`. If a dict is passed and a response with the same status
                code is already present, the existing data will be overwritten.
            deprecated: Flag this endpoint as deprecated in API docs.
            hide: Hide this endpoint in API docs.
            operation_id: The `operationId` of this endpoint. Set config `AUTO_OPERATION_ID` to
                `True` to enable the auto-generating of operationId (in the format of
                `{method}_{endpoint}`).
            security: The `security` used for this endpoint. Match the security info specified in
                the `SECURITY_SCHEMES` configuration. If you don't need specify the scopes, just
                pass a security name (equals to `[{'foo': []}]`) or a list of security names (equals
                to `[{'foo': []}, {'bar': []}]`).
            extensions: The spec extensions of this endpoint (OpenAPI operation object). The fields
                in this extensions dict should start with "x-" prefix. See more details in the
                [Specification Extensions](https://spec.openapis.org/oas/v3.1.0#specification-extensions)
                chapter of OpenAPI docs.

        *Version changed: 2.2.0*

        - Add `extensions` parameter to support setting spec extensions.

        *Version changed: 2.0.0*

        - Remove the deprecated `tag` parameter.
        - Expand `responses` to support additional structure and parameters.

        *Version changed: 1.0*

        - Add `security` parameter to support customizing security info.
        - The `role` parameter is deprecated.

        *Version changed: 0.12.0*

        - Move to `APIFlask` and `APIBlueprint` classes.

        *Version changed: 0.10.0*

        - Add parameter `operation_id`.

        *Version changed: 0.5.0*

        - Change the default value of parameters `hide` and `deprecated` from `False` to `None`.

        *Version changed: 0.4.0*

        - Add parameter `tag`.

        *Version changed: 0.3.0*

        - Change the default value of `deprecated` from `None` to `False`.
        - Rename parameter `tags` to `tag`.

        *Version added: 0.2.0*
        """

        def decorator(f):
            f = _ensure_sync(f)
            _annotate(
                f,
                summary=summary,
                description=description,
                tags=tags,
                responses=responses,
                deprecated=deprecated,
                hide=hide,
                operation_id=operation_id,
                security=security,
                extensions=extensions,
            )
            return f

        return decorator
